# import machine_learning.neural_networks as neural_networks
# import machine_learning.decision_trees as decision_trees
# import machine_learning.agnostic as agnostic
# import machine_learning.library as library
# from ._linear import *
# from ._custom_regressor import *
# from ._image import *
# from ._utilities import *
# from machine_learning._templates import load as load
# __version__ = agnostic.__version__

from . import neural_networks
from .utilities import split, kfold, MSE, RMSE, R2, accuracy, plot_correlation, plot_confusion_matrix
__version__ = "0.0.0"



import torch
import torch.nn.functional as F
from typing import Union, List, Tuple
from .pooling_stage import PoolingStage2d, PoolingStage1d, load_poolingstage


class _UNet(torch.nn.Module):
    """
    An UNet structure is a encoder followed by a decoder
    with a skip connection between layers of same depth
    """

    @classmethod
    def from_dump(cls, dump: dict) -> '_UNet':
        obj = cls(dump["in channels"])
        obj.upward = torch.nn.ModuleList()
        for d in dump["upward"]:
            pooling_stage = load_poolingstage(d)
            obj.upward.append(pooling_stage)
        obj.downward = torch.nn.ModuleList()
        for d in dump["downward"]:
            pooling_stage = load_poolingstage(d)
            obj.downward.append(pooling_stage)
        return obj

    def __init__(self, PoolingStage: type,
                 in_channels: int,
                 downward: List[Union[dict, List[dict]]],
                 pooling: List[int],
                 upward: List[Union[dict, List[dict]]],
                 pooling_type: str = "Max",
                 padded: bool = True,
                 activation: str = "relu"):
        super().__init__()
        assert len(downward) == len(upward) > 0
        self.in_channels = in_channels
        self.downward = torch.nn.ModuleList()
        downward_channels = []
        for down, pool in zip(downward, pooling):
            downward_channels.append(in_channels)
            cp = PoolingStage(in_channels, down,
                              pooling_type=pooling_type,
                              pooling_size=pool,
                              padded=padded,
                              activation=activation)
            self.downward.append(cp)
            in_channels = cp.out_channels
        self.upward = torch.nn.ModuleList()
        for up, down_channels in zip(upward, downward_channels[::-1]):
            cp = PoolingStage(in_channels+down_channels, up,
                              pooling_type=None,
                              padded=True,
                              activation=activation)
            self.upward.append(cp)
            in_channels = cp.out_channels

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        results = []
        for stage in self.downward:
            results.append(X)
            X = stage(X)
        for upward, downward, Y in zip(self.upward, self.downward[::-1],
                                       results[::-1]):
            X = self.align(Y, downward.upsample(X))
            X = upward(X)
        return X

    def shape_out(self, shape_in: list) -> list:
        shape = shape_in
        for stage in self.upward:
            shape = stage.shape_out(shape)
        return shape

    def shape_in(self, shape_out: list) -> list:
        shape = shape_out
        for stage in self.stages[::-1]:
            shape = stage.shape_in(shape)
        return shape

    @property
    def out_channels(self):
        if len(self.upward) > 0:
            return self.upward[-1].out_channels
        else:
            return self.in_channels

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "in channels": self.in_channels,
                "downward": [d.dump for d in self.downward],
                "upward": [d.dump for d in self.upward]}


class UNet1d(_UNet):

    def __init__(self, *args, **kwargs):
        super().__init__(PoolingStage1d, *args, **kwargs)

    def align(self, X1: torch.Tensor, X2: torch.Tensor):
        delta = X1.shape[2]-X2.shape[2]
        if delta != 0:
            X2 = F.pad(X2, (0, delta), mode="constant", value=0.)
        return torch.cat([X1, X2], dim=1)


class UNet2d(_UNet):

    def __init__(self, *args, **kwargs):
        super().__init__(PoolingStage2d, *args, **kwargs)

    def align(self, X1: torch.Tensor, X2: torch.Tensor):
        dh = X1.shape[2]-X2.shape[2]
        dw = X1.shape[3]-X2.shape[3]
        if (dh, dw) != (0, 0):
            X2 = F.pad(X2, (0, dw, 0, dh), mode="constant", value=0.)
        return torch.cat([X1, X2], dim=1)



import math
import copy
import torch
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from typing import Union, Callable, Type, List, Iterable
from collections import OrderedDict
from .conversions import tensor_to_index, tensor_to_probabilities


def neural_network(cls: torch.nn.Module) -> Type:
    """A decorator that wraps a 'torch.nn.Module' to hide complexity"""

    class NeuralNetwork:
        """
        Parameters
        ----------
        module : torch.nn.Module
            the underlying torch module of the model
        optimizer : torch.optim
            the optimizer used for training
        """
        def __init__(self, *args,
                     GPU: bool = False,
                     learning_rate: float = 1.0E-3,
                     norm_update_factor: Union[float, None] = 0.1,
                     optimization_method: str = "Adam",
                     L1: Union[float, None] = None,
                     L2: Union[float, None] = None,
                     **kwargs):
            """
            Parameters
            ----------
            *args : tuple
                The args passed to the constructor of 'self.module'
            GPU : bool
                If True, calculation on GPU is enabled
            learning_rate : float
                The learning rate of the model
            norm_update_factor : float or None
                The update factor used for batch normalization
            optimization_method : str
                The name of the optimization method
            L1 : float or None
                L1 regularization added to the loss function
            L2 : float or None
                L2 regularization added to the loss function
            **kwargs : dict
                The kwargs passed to the constructor of 'self.module'
            """
            self.module = cls(*args, **kwargs)
            self.GPU = GPU
            self.optimization_method = optimization_method
            self.learning_rate = learning_rate
            self.norm_update_factor = norm_update_factor
            self.L1 = L1
            self.L2 = L2
            self.residuals = {"training loss": [],
                              "validation loss": [],
                              "epochs": [],
                              "best epoch": None}

        def fit(self, training_data: Union[tuple, Callable],
                validation_data: Union[tuple, Callable, None] = None,
                n_epochs: int = 1000,
                patience: int = 100,
                verbose: bool = True,
                L_minibatchs: Union[int, None] = None):
            """
            Trains a neural network model.

            Parameters
            ----------
            training_data : tuple or callable
                The data used to fit the model on.
                A tuple of (x, y[, weights]) or a callable that yields them.
                The type of each element depends on the model kind.
            validation_data : tuple or callable or None
                The data used to test for early stoping.
                Similar to training_data or None
            n_epochs : int
                The maximum number of epochs
            patience : int
                The number of epochs before early stopping
            verbose : bool
                If True the loss are displayed at each epoch
            L_minibatchs : int or None
                Maximum size of the minibatch
                Or None to process the full batch in one go
            """
            self.module.train()
            # Converts training/validation data to tensors
            if not callable(training_data):
                training_data = self.module.data_to_tensor(*training_data)
            if not callable(validation_data) and validation_data is not None:
                validation_data = self.module.data_to_tensor(*validation_data)
            # Initializing
            if self.residuals["best epoch"] is None:
                best_loss = float("inf")
                best_epoch = 0
            else:
                best_epoch = self.residuals["best epoch"]
                i = self.residuals["epochs"].index(best_epoch)
                best_loss = self.residuals["validation loss"][i]
                for v in ["validation loss", "training loss", "epochs"]:
                    self.residuals[v] = self.residuals[v][:i+1]
            best_state = self._get_state()
            # trains the model, stops if the user press 'ctrl+c'
            self._training_loop(training_data, validation_data, n_epochs,
                                patience, verbose, L_minibatchs, best_epoch,
                                best_loss, best_state)

        def plot_residuals(self, ax=None, log: bool = True):
            """
            Plot the training and validation data residuals

            Parameters
            ----------
            ax : matplotlib.axes.Axes or None
                The axes to plot on
            log : bool
                If true the y axis is in log scale
            """
            if ax is None:
                f, ax = plt.subplots()
            epochs = self.residuals["epochs"]
            ax.scatter(epochs, self.residuals["training loss"],
                       marker=".",
                       label="training loss")
            ax.scatter(epochs, self.residuals["validation loss"],
                       marker=".",
                       label="validation loss")
            if self.residuals["best epoch"] is not None:
                ax.axvline(self.residuals["best epoch"], color="k")
            if log:
                ax.set_yscale("log")
            ax.set_ylabel("loss")
            ax.set_xlabel("epochs")
            ax.legend()

        def __call__(self, X):
            """
            Returns the model's evaluation on the given input

            Parameters:
            -----------
            X : Any
                The input X of the model.
                it's type depend on the neural network type.
                see 'help(self.module)'

            Returns
            -------
            Any :
                The returned Y value of the model.
                it's type depends on the neural network type.
                see 'help(self.module)'
            """
            self.module.eval()
            x, _, _ = self.module.data_to_tensor(X, None, None)
            y = self.module(x)
            return self.module.tensor_to_y(y)

        @property
        def GPU(self) -> bool:
            """
            Returns whether the model is evaluated/trained on GPU

            Returns
            -------
            bool :
                whether the GPU mode is enabled
            """
            return self._GPU

        @GPU.setter
        def GPU(self, enable: bool):
            """
            Set whether the model should be evaluated/trained on GPU.
            This is only available with a CUDA capable GPU.

            Returns
            -------
            bool :
                whether the GPU mode is enabled
            """
            if hasattr(self, "_GPU") and self._GPU == enable:
                return
            if enable and not(torch.cuda.is_available()):
                raise RuntimeError("CUDA is not available on this computer, \
                                    can't train on GPU")
            self._GPU = enable
            if self.GPU:
                self.module.device = torch.device("cuda:0")
            else:
                self.module.device = torch.device("cpu")
            self.module.to(self.module.device)

        @property
        def learning_rate(self) -> float:
            """
            Returns the learning rate used for training

            Returns
            -------
            float :
                the learning rate
            """
            if hasattr(self, "_learning_rate"):
                return self._learning_rate
            else:
                return 0

        @learning_rate.setter
        def learning_rate(self, lr: float):
            """
            set the learning rate for the training

            Parameters:
            -----------
            lr : float
                new learning rate
            """
            self._learning_rate = lr
            for g in self.optimizer.param_groups:
                g["lr"] = lr

        @property
        def optimization_method(self) -> str:
            """
            returns the name of the optimization method

            Returns
            -------
            str :
                name of the method
            """
            return self._optimization_method

        @optimization_method.setter
        def optimization_method(self, name: str):
            """
            set the optimization method for training the model.
            must be the name of an optimizer class from 'torch.optim'.

            This also resets the optimization parameters (gradient momentum,
            learning rate decay, ...)

            Parameters
            ----------
            name : str
                the name of the optimization method
            """
            if not hasattr(torch.optim, name):
                available = [n for n in dir(torch.optim) if n[0] != "_"]
                raise ValueError(f"Invalid optimizer '{name}', "
                                 f"valid options are: {available}")
            cls = getattr(torch.optim, name)
            self.optimizer = cls(self.module.parameters(), self.learning_rate)
            self._optimization_method = name

        @property
        def norm_update_factor(self):
            """
            Returns the update factor used for batch normalization

            Returns
            -------
            float or None:
                the update factor
            """
            return self._f

        @norm_update_factor.setter
        def norm_update_factor(self, f: Union[float, None]):
            """
            Set the update factor 'f' used for batch normalization
            moment = f*batch_moment + (1-f)*moment
            Where 'moment' are the mean and variance

            f must be between 0. and 1.
            or None to use an averaging method instead

            Parameters
            ----------
            f : float or None
                the update factor
            """
            assert (f is None) or (0. <= f <= 1.)
            for m in self.module.modules():
                if type(m).__name__.startswith("BatchNorm"):
                    m.momentum = f
            self._f = f

        def _get_state(self) -> tuple:
            """
            Returns a snapshot of the model's state

            The 'state_dict' are deep copied otherwise the saved tensors are
            modified along with the network's training

            Returns
            -------
            tuple :
                the tuple of (module, grad, optimizer) state dicts
            """
            module_state = copy.deepcopy(self.module.state_dict())
            params = self.module.state_dict(keep_vars=True)
            grad_state = OrderedDict([(key, self._copy_grad(params[key].grad))
                                      for key in params])
            optimizer_state = copy.deepcopy(self.optimizer.state_dict())
            return module_state, grad_state, optimizer_state

        def _copy_grad(self, grad: Union[torch.Tensor, None]
                       ) -> Union[torch.Tensor, None]:
            """returns a copy of the given grad tensor"""
            return grad if grad is None else grad.detach().clone()

        def _set_state(self, state: tuple):
            """
            Loads a snapshot of the model's state, as returned by '_get_state'

            Parameters
            ----------
            state : tuple
                The tuple of (module.state_dict(), optimizer.state_dict())
            """
            module_state, grad_state, optimizer_state = state
            self.module.load_state_dict(module_state)
            state = self.module.state_dict(keep_vars=True)
            for key in module_state.keys():
                state[key].grad = grad_state[key]
            self.optimizer.load_state_dict(optimizer_state)

        def _training_loop(self, training_data: tuple,
                           validation_data: Union[tuple, None],
                           n_epochs: int,
                           patience: int,
                           verbose: bool,
                           L_minibatchs: Union[int, None],
                           best_epoch: int,
                           best_loss: float,
                           best_state: tuple):
            """
            Trains the model for a fixed number of epoch,
            or until validation loss has'nt decreased for 'patience' epochs,
            or until the user press 'ctrl+c'

            At each epoch:
                The parameters are updated using the gradient (0 initially)
                The gradient is set back to 0 (otherwise it accumulates)
                The gradient of the loss is evaluated on the training data
                if validation data are provided:
                    The validation loss is evaluated
                    if the validation loss is inferior to the previous best:
                        the best loss is updated
                        the state of the model is saved
                    otherwise if we waited more than 'patience' epochs:
                        interupts training
            The best state and epoch are then saved

            Parameters:
            -----------
            training_data : tuple or Callable
                The data provided to 'self._batch'
            validation_data : tuple or Callbale or None
                Same as 'training_data' or None if not using early stoping
            n_epochs : int
                The number of epochs to performe
            patience : int
                The number of epochs waiting for improvement
                before early stoping
            verbose : bool
                If True prints models info while training
            L_minibatchs : int or None
                The maximum number of items in a minibatchs
                or None to not to use minibatchs
            best_epoch : int
                the epoch of the previous best state
            best_loss : float
                the value of the previous best validation loss
            best_state : tuple
                the snapshot of the model as returned by 'self._get_state'
            """
            try:
                for epoch in range(best_epoch+1, best_epoch+n_epochs+1):
                    self.optimizer.step()
                    self.optimizer.zero_grad()
                    training_loss = self._batch(training_data, L_minibatchs,
                                                train=True)
                    if validation_data is not None:
                        validation_loss = self._batch(validation_data, None,
                                                      train=False)
                        if validation_loss < best_loss:
                            best_epoch = epoch
                            best_loss = validation_loss
                            best_state = self._get_state()
                        elif (epoch - best_epoch) > patience:
                            break
                    else:
                        best_epoch = epoch
                        validation_loss = None
                    self.residuals["training loss"].append(training_loss)
                    self.residuals["validation loss"].append(validation_loss)
                    self.residuals["epochs"].append(epoch)
                    if verbose:
                        msg = f"Epoch {epoch}: train={training_loss:.3g}"
                        if validation_loss is not None:
                            msg += f" val={validation_loss:.3g}"
                            if best_epoch != epoch:
                                msg += " - no improvement"
                        print(msg, flush=True)
            except KeyboardInterrupt:
                if verbose:
                    print("Training interupted by the user", flush=True)
                # Trims data in case user interupted in the midle of the loop
                keys = ["validation loss", "training loss", "epochs"]
                L = min(len(self.residuals[key]) for key in keys)
                for key in keys:
                    self.residuals[key] = self.residuals[key][:L]
                best_epoch = min(self.residuals["epochs"][-1], best_epoch)
            finally:
                # load the best state
                if validation_data is not None:
                    self._set_state(best_state)
                # Save the best epoch
                self.residuals["best epoch"] = best_epoch

        def _batch(self, data: Union[tuple, Callable],
                   *args, **kwargs) -> float:
            """
            Process the data by batch (or in one go),
            and returns the average loss.

            Parameters
            ----------
            data : tuple or Callable
                The (X, Y, weights) to evaluate the loss on,
                or a function that yields them by batch.
                'X' and 'Y' are tensors, 'weights' is a list of float or None.
            *args : tuple
                The arguments passed to _minibatch
            **kwargs : dict
                The keyword arguments passed to _minibatch

            Returns
            -------
            float :
                The loss function averaged over the batchs
            """
            if not callable(data):
                return self._minibatch(data, *args, **kwargs)
            else:
                losses = []
                for batch_data in data():
                    loss = self._minibatch(self.data_to_tensor(*batch_data),
                                           *args, **kwargs)
                    losses.append(loss)
                return sum(losses)/len(losses)

        def _minibatch(self, batch_data: tuple,
                       L_minibatchs: Union[int, None],
                       train: bool = False) -> float:
            """
            Process the input batch data in minibatchs
            of maximum size 'L_minibatch', and returns the average loss.

            Parameters
            ----------
            batch_data : tuple
                The (X, Y, weights) to evaluate the loss on.
                'X' and 'Y' are tensors, 'weights' is a list of float or None.
            L_minibatch : int or None
                The maximum number of observations in a minibatch.
                If None the whole data is processed in one go.
            train : bool
                If True, the gradient is back propagated

            Returns
            -------
            float :
                The loss function averaged over the minibatchs
            """
            if L_minibatchs is None:
                return self._eval_loss(*batch_data, train=train)
            else:
                X, Y, weights = batch_data
                L_minibatchs = min(max(1, L_minibatchs), len(X))
                n = math.ceil(len(X)/L_minibatchs)
                bounds = [int(i*len(X)/n) for i in range(n+1)]
                losses = []
                for (start, end) in zip(bounds[:-1], bounds[1:]):
                    x = X[start:end]
                    y = Y[start:end]
                    w = None if weights is None else weights[start:end]
                    losses.append(self._eval_loss(x, y, w, train=train))
                return sum(losses)/len(losses)

        def _eval_loss(self, x: torch.Tensor, y: torch.Tensor,
                       w: Union[List[float], None] = None,
                       train: bool = False) -> torch.Tensor:
            """
            Evaluates the loss on the given input data.

            If 'train' is True, builds the computational graph
            (Slower to compute/more memory used)
            and backpropagate the gradient.

            Parameters
            ----------
            x : torch.Tensor
                observations
            y : torch.Tensor
                target
            w : List of float, or None
                weights
            train : bool
                If True, grad is backpropagated

            Returns
            -------
            float :
                scalar tensor of the evaluated loss
            """
            if train:
                loss = self.module.loss(self.module(x), y, w)
                loss = self._regularization(loss)
                loss.backward()
                return float(loss)
            else:
                with torch.no_grad():
                    self.module.eval()
                    loss = self.module.loss(self.module(x), y, w)
                    self.module.train()
                    loss = self._regularization(loss)
                    return float(loss)

        def _regularization(self, loss: torch.Tensor) -> torch.Tensor:
            """
            Add L1 and L2 regularization terms to the loss

            Parameters
            ----------
            loss : torch.Tensor
                the scalar tensor representing the loss

            Returns
            -------
            torch.Tensor :
                the regularized loss
            """
            if self.L1 is not None:
                norm = sum([torch.norm(p, 1)
                            for p in self.module.parameters()])
                loss = loss + self.L1 * norm
            if self.L2 is not None:
                norm = sum([torch.norm(p, 2)
                            for p in self.module.parameters()])
                loss = loss + self.L2 * norm
            return loss

    return NeuralNetwork


def nn_classifier(cls: Type) -> Type:
    """A decorator that wraps a classifier 'torch.nn.Module'"""

    class NeuralNetworkClassifier(neural_network(cls)):

        def __init__(self, *args, class_weights=None, **kwargs):
            super().__init__(*args, **kwargs)
            self.class_weights = class_weights

        def index(self, X) -> np.ndarray:
            """
            Returns the category index for each observation

            Parameters
            ----------
            X : Any
                The input X of the model.
                it's type depend on the neural network type.
                see 'help(self.module)'

            Returns
            -------
            np.ndarray :
                An array of long, indexes of the 'self.category'
            """
            x, _, _ = self.module.data_to_tensor(X, None, None)
            return tensor_to_index(self.module(x))

        def probability(self, X) -> pd.DataFrame:
            """
            Return the category probabilities for each observations

            Parameters
            ----------
            X : Any
                The input X of the model.
                it's type depend on the neural network type.
                see 'help(self.module)'

            Returns
            -------
            pd.DataFrame :
                DataFrame of category probabilities
                where each column corresponds to a category
            """
            x, _, _ = self.module.data_to_tensor(X, None, None)
            return tensor_to_probabilities(self.module(x))

        @property
        def class_weights(self):
            if self.module.class_weights is None:
                return None
            else:
                return self.module.class_weights.tolist()

        @class_weights.setter
        def class_weights(self, other: Iterable[float]):
            if other is not None:
                assert len(other) == len(self.categories)
                other = torch.tensor(other, dtype=torch.float,
                                     device=self.device)
            self.module.class_weights = other

        @property
        def categories(self):
            return self.module.categories

    return NeuralNetworkClassifier



import torch


class Linear(torch.nn.Linear):
    """A wrapper around torch.nn.Linear"""

    @classmethod
    def from_dump(cls, dump: dict) -> 'Linear':
        """returns a 'Linear' layer from a dump"""
        assert dump["type"] == cls.__name__
        obj = cls(dump["in features"],
                  dump["out features"],
                  bias=(dump["bias"] is not None))
        obj.weight = torch.nn.Parameter(torch.tensor(dump["weight"],
                                                     dtype=torch.float))
        if obj.bias is not None:
            obj.bias = torch.nn.Parameter(torch.tensor(dump["bias"],
                                                       dtype=torch.float))
        return obj

    def __init__(self, in_features: int, out_features: int, bias=True):
        """
        Parameters
        ----------
        in_features : int
            The number of input features
        out_features : int
            The number of output features
        bias : bool
            If False, there is not bias term in the linear equation:
            y = weight*x + bias
        """
        super().__init__(in_features, out_features, bias=bias)

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "in features": self.in_features,
                "out features": self.out_features,
                "weight": self.weight.tolist(),
                "bias": self.bias if self.bias is None else self.bias.tolist()}



import torch
from typing import Tuple


def load_convolution(dump: dict):
    """returns the Conv1d or Conv2d object loaded from a dump"""
    return globals()[dump["type"]].from_dump(dump)


class _Convolution:
    """A template for batch norm layers"""

    @classmethod
    def from_dump(cls, dump: dict) -> '_Convolution':
        """returns a '_Convolution' layer from a dump"""
        assert dump["type"] == cls.__name__
        obj = cls(dump["in channels"],
                  dump["out channels"],
                  dump["kernel size"],
                  stride=dump["stride"],
                  with_bias=dump["with bias"])
        obj.weight = torch.nn.Parameter(
                        torch.tensor(dump["weight"], dtype=torch.float))
        if obj.with_bias:
            obj.bias = torch.nn.Parameter(
                        torch.tensor(dump["bias"], dtype=torch.float))
        return obj

    @property
    def dump(self) -> dict:
        d = {"type": type(self).__name__,
             "in channels": self.in_channels,
             "out channels": self.out_channels,
             "kernel size": list(self.kernel_size),
             "stride": list(self.stride),
             "with bias": self.with_bias,
             "weight": self.weight.tolist()}
        if self.with_bias:
            d["bias"] = self.bias.tolist()
        return d

    def shape_out(self, shape_in: list) -> list:
        assert shape_in[0] == self.in_channels
        dims = [int((d - k)/s + 1) for d, k, s in
                zip(shape_in[1:], self.kernel_size, self.stride)]
        return [self.out_channels] + dims

    def shape_in(self, shape_out: list) -> list:
        assert shape_out[0] == self.out_channels
        dims = [(d-1)*s + k for d, k, s in
                zip(shape_out[1:], self.kernel_size, self.stride)]
        return [self.in_channels] + dims


class Conv1d(torch.nn.Conv1d, _Convolution):
    """A wrapper around torch.nn.Conv1d"""

    def __init__(self, in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 stride: int = 1,
                 with_bias: bool = True):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        out_channels : int
            The number of output channels
        kernel_size : int
            The size of the convolved kernel
        stride : int
            The stride used for the convolution
        with_bias : bool
            Whether to have a bias in the linear equation:
            y = weight*x + bias
        """
        self.with_bias = with_bias
        torch.nn.Conv1d.__init__(self, in_channels, out_channels,
                                 kernel_size, stride, bias=with_bias)


class Conv2d(torch.nn.Conv2d, _Convolution):
    """A wrapper around torch.nn.Conv2d"""

    def __init__(self, in_channels: int,
                 out_channels: int,
                 kernel_size: Tuple[int, int],
                 stride: Tuple[int, int] = (1, 1),
                 with_bias: bool = True):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        out_channels : int
            The number of output channels
        kernel_size : tuple of int
            The (height, width) size of the convolved kernel
        stride : tuple of int
            The (height, width) stride used for the convolution
        with_bias : bool
            Whether to have a bias in the linear equation:
            y = weight*x + bias
        """
        self.with_bias = with_bias
        torch.nn.Conv2d.__init__(self, in_channels, out_channels,
                                 kernel_size, stride, bias=with_bias)



from .._library_loader import _lib
from ._tensor import Tensor as _Tensor
import ctypes as _c

# Typing of the layers functions
_lib.pad.restype = _c.c_void_p
_lib.pad.argtypes = [_c.c_void_p, _c.c_double,
                     _c.c_uint, _c.c_uint,
                     _c.c_uint, _c.c_uint]

_lib.convolve.restype = _c.c_void_p
_lib.convolve.argtypes = [_c.c_void_p, _c.c_void_p, _c.c_void_p,
                          _c.c_uint, _c.c_uint]

_lib.max_pool.restype = _c.c_void_p
_lib.max_pool.argtypes = [_c.c_void_p,
                          _c.c_uint, _c.c_uint]

_lib.batch_normalize.restype = _c.c_void_p
_lib.batch_normalize.argtypes = [_c.c_void_p, _c.c_void_p,
                                 _c.c_void_p, _c.c_void_p, _c.c_void_p]

_lib.linear.restype = _c.c_void_p
_lib.linear.argtypes = [_c.c_void_p, _c.c_void_p, _c.c_void_p]

_lib.resample.restype = _c.c_void_p
_lib.resample.argtypes = [_c.c_void_p, _c.c_uint, _c.c_uint]

_lib.Tensor_tanh.restype = _c.c_void_p
_lib.Tensor_tanh.argtypes = [_c.c_void_p]

_lib.Tensor_relu.restype = _c.c_void_p
_lib.Tensor_relu.argtypes = [_c.c_void_p]


# defining the functions
def pad(input: _Tensor, value: float, size: tuple):
    assert isinstance(input, _Tensor) and (input.ndims == 3)
    left, right, top, bottom = size
    return _Tensor(_c.c_void_p(_lib.pad(input.pointer, value, left,
                                        right, top, bottom)))


def convolve(input: _Tensor, kernel: _Tensor, bias: _Tensor, stride: tuple):
    assert isinstance(input, _Tensor) and (input.ndims == 3)
    assert isinstance(kernel, _Tensor) and (kernel.ndims == 4)
    assert isinstance(bias, _Tensor) and (bias.ndims == 1)
    si, sk = input.shape[-2:], kernel.shape[-2:]
    assert all([k <= i for i, k in zip(si, sk)])
    S_h, S_w = stride
    return _Tensor(_c.c_void_p(_lib.convolve(input.pointer,
                                             kernel.pointer,
                                             bias.pointer,
                                             S_h, S_w)))


def max_pool(input: _Tensor, window: tuple):
    assert isinstance(input, _Tensor) and (input.ndims == 3)
    height, width = window
    return _Tensor(_c.c_void_p(_lib.max_pool(input.pointer, height, width)))


def batch_normalize(input: _Tensor, mean: _Tensor, variance: _Tensor,
                    weight: _Tensor, bias: _Tensor):
    assert isinstance(input, _Tensor) and (input.ndims == 3)
    assert isinstance(mean, _Tensor) and (mean.ndims == 1)
    assert isinstance(variance, _Tensor) and (variance.ndims == 1)
    assert isinstance(weight, _Tensor) and (weight.ndims == 1)
    assert isinstance(bias, _Tensor) and (bias.ndims == 1)
    return _Tensor(_c.c_void_p(_lib.batch_normalize(input.pointer,
                                                    mean.pointer,
                                                    variance.pointer,
                                                    weight.pointer,
                                                    bias.pointer)))


def linear(input: _Tensor, weight: _Tensor, bias: _Tensor):
    assert isinstance(input, _Tensor) and (input.ndims == 1)
    assert isinstance(weight, _Tensor) and (weight.ndims == 2)
    assert isinstance(bias, _Tensor) and (bias.ndims == 1)
    return _Tensor(_c.c_void_p(_lib.linear(input.pointer,
                                           weight.pointer,
                                           bias.pointer)))


def resample(input: _Tensor, new_height: int, new_width: int) -> _Tensor:
    assert (new_height > 0) and (new_width > 0)
    return _Tensor(_c.c_void_p(_lib.resample(input.pointer,
                                             new_height,
                                             new_width)))


def tanh(input: _Tensor):
    assert isinstance(input, _Tensor)
    return _Tensor(_c.c_void_p(_lib.Tensor_tanh(input.pointer)))


def relu(input: _Tensor):
    assert isinstance(input, _Tensor)
    return _Tensor(_c.c_void_p(_lib.Tensor_relu(input.pointer)))


def identity(input: _Tensor) -> _Tensor:
    return input


# class Linear:
#     def __init__(self, parameters: dict):
#         self.weight = _Tensor(parameters["weights"])
#         if parameters["bias"] is None:
#             self.bias = _Tensor([0.]*self.weights.shape[0])
#         else:
#             self.bias = _Tensor(parameters["bias"])

#     def __call__(self, input) -> _Tensor:
#         return linear(input, self.weight, self.bias)


# class Padding:
#     def __init__(self, parameters: dict):
#         self.value = parameters["value"]
#         self.size = parameters["size"]

#     def __call__(self, input: _Tensor) -> _Tensor:
#         return pad(input, self.value, self.size)


# class Convolution:
#     def __init__(self, parameters: dict):
#         self.weights = _Tensor(parameters["weights"])
#         if parameters["bias"] is None:
#             self.bias = _Tensor([0]*self.weights.shape[0])
#         else:
#             self.bias = _Tensor(parameters["bias"])
#         self.stride = parameters["stride"]

#     def __call__(self, input: _Tensor) -> _Tensor:
#         return convolve(input, self.weights, self.bias, self.stride)


# class BatchNorm:
#     def __init__(self, parameters: dict):
#         self.mean = _Tensor(parameters["running_mean"])
#         self.variance = _Tensor(parameters["running_var"])
#         self.weight = _Tensor(parameters["weight"])
#         self.bias = _Tensor(parameters["bias"])

#     def __call__(self, input: _Tensor) -> _Tensor:
#         return batch_normalize(input, self.mean, self.variance,
#                                self.weight, self.bias)


# class MaxPool:
#     def __init__(self, parameters: dict):
#         self.window = parameters["window"]

#     def __call__(self, input: _Tensor) -> _Tensor:
#         return max_pool(input, self.window)


# class NonLinear:
#     def __init__(self, name: str):
#         if name is None:
#             name = "identity"
#         module = _sys.modules[__name__]
#         if not hasattr(module, name):
#             raise ValueError(f"non-linear function '{name}' "
#                              "is not implemented")
#         self.func = getattr(module, name)

#     def __call__(self, input: _Tensor) -> _Tensor:
#         return self.func(input)


# class ConvolutionLayer(_m.Model):
#     def __init__(self, dump):
#         super().__init__(dump)
#         self.padding = Padding(self.parameters["padding"])
#         self.convolution = Convolution(self.parameters["convolution"])
#         self.batch_norm = BatchNorm(self.parameters["batch_norm"])
#         self.pooling = MaxPool(self.parameters["pooling"])
#         self.non_linear = NonLinear(self.parameters["non_linear"])

#     def __call__(self, input: _Tensor) -> _Tensor:
#         T = self.padding(input)
#         T = self.convolution(T)
#         T = self.batch_norm(T)
#         T = self.pooling(T)
#         return self.non_linear(T)


# class DenseLayer(_m.Model):
#     def __init__(self, dump):
#         super().__init__(dump)
#         self.linear = Linear(self.parameters["linear"])
#         self.non_linear = NonLinear(self.parameters["non_linear"])

#     def __call__(self, input: _Tensor) -> _Tensor:
#         T = self.linear(input)
#         return self.non_linear(T)



import torch
from typing import Union, List, Tuple
from .convolution_stage import ConvStage1d, ConvStage2d, \
                                load_activatedconv
from .operations import linear_interpolation, bilinear_interpolation
from . import pooling as pool


def load_poolingstage(dump: dict):
    return globals()[dump["type"]].from_dump(dump)


class _PoolingStage(torch.nn.Module):
    """
    A succession of activated convolutions
    followed by an optional pooling operation
    """

    @classmethod
    def from_dump(cls, dump):
        assert dump["type"] == cls.__name__
        obj = cls(dump["in channels"])
        obj.pooled = dump["pooled"]
        obj.convolutions = torch.nn.ModuleList()
        obj.convolutions.extend([load_activatedconv(d)
                                 for d in dump["convolutions"]])
        obj.pooling = pool.load_pooling(dump["pooling"])
        return obj

    def __init__(self):
        super().__init__()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        X = input
        for conv in self.convolutions:
            X = conv(X)
        if self.pooled:
            X = self.pooling(X)
        return X

    def shape_in(self, shape_out: list) -> list:
        shape = self.pooling.shape_in(shape_out)
        for conv in self.convolutions[::-1]:
            shape = conv.shape_in(shape)
        return shape

    def shape_out(self, shape_in: list) -> list:
        shape = shape_in
        for conv in self.convolutions:
            shape = conv.shape_out(shape)
        return self.pooling.shape_out(shape)

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "in channels": self.in_channels,
                "pooled": self.pooled,
                "convolutions": [c.dump for c in self.convolutions],
                "pooling": self.pooling.dump}

    @property
    def out_channels(self):
        if len(self.convolutions) == 0:
            return self.in_channels
        return self.convolutions[-1].out_channels


class PoolingStage1d(_PoolingStage):

    def __init__(self, in_channels,
                 convolutions: Union[dict, List[dict]] = dict(),
                 pooling_type: Union[None, str] = None,
                 pooling_size: int = 2,
                 padded: bool = True,
                 stacked: bool = False,
                 activation: str = "relu"):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        convolutions : list of dict or list of list of dict
            The {"channels", "window", "stride", "bias", "stacked"} kwargs
            used to create each convolution
        pooling_type : one of {"Max", "Avg"}
            The type of pooling to perform after each convolution stage
        pooling_size : tuple of int
            the (height, width) of the pooling window
        default_padded : bool
            If "padded" is omited in a convolution kwarg
            use this value as default
        stacked: bool
            default value for the "stacked" key in the kwargs
            If omited, the default is used
        default_activation : str
            if "activation" is omited in a convolution kwarg
            use this as default
        """
        super().__init__()
        if isinstance(convolutions, dict):
            convolutions = [convolutions]
        self.in_channels = in_channels
        self.pooled = (pooling_type is not None)
        if pooling_type is None:
            pooling_type = "Max"
            pooling_size = 1
        self.convolutions = torch.nn.ModuleList()
        for kwargs in convolutions:
            kwargs.setdefault("padded", padded)
            kwargs.setdefault("stacked", stacked)
            kwargs.setdefault("activation", activation)
            conv = ConvStage1d(in_channels, **kwargs)
            self.convolutions.append(conv)
            in_channels = conv.out_channels
        cls = getattr(pool, pooling_type+"Pool1d")
        self.pooling = cls(kernel_size=pooling_size, stride=pooling_size)

    def upsample(self, X):
        L = X.shape[2]
        _, h, w = self.shape_in([self.out_channels, L])
        return linear_interpolation(X, size=(h, w))


class PoolingStage2d(_PoolingStage):

    def __init__(self, in_channels,
                 convolutions: Union[dict, List[dict]] = dict(),
                 pooling_type: Union[None, str] = None,
                 pooling_size: Tuple[int, int] = (2, 2),
                 padded: bool = True,
                 stacked: bool = False,
                 activation: str = "relu"):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        convolutions : list of dict or list of list of dict
            The {"channels", "window", "stride", "bias", "stacked"} kwargs
            used to create each convolution
        pooling_type : one of {"Max", "Avg"}
            The type of pooling to perform after each convolution stage
        pooling_size : tuple of int
            the (height, width) of the pooling window
        padded : bool
            If "padded" is omited in a convolution kwarg
            use this value as default
        stacked: bool
            default value for the "stacked" key in the kwargs
            If omited, the default is used
        activation : str
            if "activation" is omited in a convolution kwarg
            use this as default
        """
        super().__init__()
        if isinstance(convolutions, dict):
            convolutions = [convolutions]
        self.in_channels = in_channels
        self.pooled = (pooling_type is not None)
        if pooling_type is None:
            pooling_type = "Max"
            pooling_size = (1, 1)
        self.convolutions = torch.nn.ModuleList()
        for kwargs in convolutions:
            kwargs.setdefault("padded", padded)
            kwargs.setdefault("stacked", stacked)
            kwargs.setdefault("activation", activation)
            conv = ConvStage2d(in_channels, **kwargs)
            self.convolutions.append(conv)
            in_channels = conv.out_channels
        cls = getattr(pool, pooling_type+"Pool2d")
        self.pooling = cls(kernel_size=pooling_size, stride=pooling_size)

    def upsample(self, X):
        h, w = X.shape[2:]
        _, h, w = self.shape_in([self.out_channels, h, w])
        return bilinear_interpolation(X, size=(h, w))



import machine_learning as _ml
import machine_learning._templates as _templates
import machine_learning.neural_networks.common as _common
import machine_learning.neural_networks.layers as _layers
import torch as _torch
import types as _types
import numpy as _np


class ImageClassifier(_common.NNtemplate, _templates.ClassifierTemplate):
    """
    A model that classify images.

    Attributes:
    -----------
    categories : list of categories
    convolution_layers : list of ConvolutionLayer
    dense_layers : list of DenseLayer
    input_shape : tuple of int
        shape of the input images formated as:
        (channels, height, width)
    args : list or None
        list of arguments used to generate the model
    """

    def __init__(self):
        super().__init__()
        self.convolution_layers = []
        self.dense_layers = []
        self.input_shape = (0, 0, 0)
        self.loss = _torch.nn.functional.cross_entropy
        self.args = None

    def fit(self, images, labels, validation=None, validation_fraction=0.2,
            windows=[], channels=[], strides=[], pooling=[], dense=[],
            l_max=None, restart=True, non_linear="relu", padded=False,
            **kwargs):
        """
        Fit the model on the given data.

        Parameters:
        -----------
        images : list of Image objects
            a list of machine_learning.Images objects
        labels : list
            a list of label object for each image
            (the labels can be a list of str for example)
        validation : tuple of (images, labels) or None
            data used for validation
        validation_fraction : float
            if validation is None, training data is splited
        windows : list of tuple of int
            the convolution windows (height, width)
        channels : list of int
            channels number at output of each convolution
        strides : list of tuple of int
            the convolution strides (y, x)
        pooling : list of tuple of int
            the polling windows (height, width)
        dense : list of int
            the dense layers
        l_max : int or None
            maximum observations in a batch
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        low_memory : bool
            If True, save memory at the cost of computation speed.
        padded : bool
            If True, a padding is applied before convolution.
            If padded is True, stride is (1, 1) and pooling is (1, 1),
            The tensor has the same shape at the output than at the input.

        Returns:
        --------
        training, validation : tuples of (images, labels)

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        training = (images, labels)
        training, validation = self._validation_split(training, validation,
                                                      validation_fraction)
        res = self._preprocess(training, validation)
        self.mean, self.std, self.input_shape, self.categories, _, _ = res
        # create layers
        if restart:
            self._set_layers(windows, channels, strides, pooling, dense,
                             non_linear, padded)
        # load data in memory
        train = self._to_tensor(training)
        val = self._to_tensor(validation)
        train, n_training = self._wrap_batch(train, l_max)
        val, n_validation = self._wrap_batch(val, l_max)
        # fitting
        self.train_loop(train, val, n_training, n_validation, **kwargs)
        # returning training and validation data
        return training, validation

    def fit_batch(self, training, validation, padded=False,
                  windows=[], channels=[], strides=[], pooling=[], dense=[],
                  restart=True, non_linear="relu", low_memory=False, **kwargs):
        """
        Fit the model on the given batches.

        Parameters:
        -----------
        training : generator function
            function that yields (images, labels)
        validation : generator function
            function that yields (images, labels)
        windows : list of tuple of int
            the convolution windows (height, width)
        channels : list of int
            channels number at output of each convolution
        strides : list of tuple of int
            the convolution strides (y, x)
        pooling : list of tuple of int
            the polling windows (height, width)
        dense : list of int
            the dense layers
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        low_memory : bool
            If True, save memory at the cost of computation speed.
        padded : bool
            If True, a padding is applied before convolution.
            If padded is True, stride is (1, 1) and pooling is (1, 1),
            The tensor has the same shape at the output than at the input.

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        self.mean, self.std, self.input_shape, self.categories, n_training, \
            n_validation = self._preprocess(training, validation)
        # create layers
        if restart:
            self._set_layers(windows, channels, strides, pooling, dense,
                             non_linear, low_memory, padded)
        # fitting
        self.train_loop(self._wrap(training), self._wrap(validation),
                        n_training, n_validation, **kwargs)

    def predict(self, images):
        """
        Make a prediction using the trained model.

        Parameters:
        -----------
        images : Image or list of Images

        Returns:
        --------
        labels : list
        """
        scalar = False
        if isinstance(images, _ml.Image):
            images = [images]
            scalar = True
        tensor = self._images_to_tensor(images, self.mean, self.std)
        with _torch.no_grad():
            output = self(tensor)
        labels = self._tensor_to_labels(output)
        if scalar:
            labels = labels[0]
        return labels

    def deep_dream(self, layer, channel, iterations=100, save_frequency=10,
                   verbose=True, input_image=None, lr=1.0E2, L2=0., L1=0.):
        """
        Returns an image representation of what best activates
        the given channel of the given layer

        https://towardsdatascience.com/how-to-visualize-convolutional-features-in-40-lines-of-code-70b7d87b0030
        """
        # create image tensor
        if input_image is None:
            data = _np.random.randint(108, 148, size=self.input_shape)
            input_image = _ml.Image(tensor=data)
        # tensor = self._images_to_tensor(input_image, mean=self.mean,
        #                                 std=self.std)
        tensor = _np.expand_dims(input_image.tensor, axis=0)
        tensor = _torch.tensor(tensor, requires_grad=True,
                               dtype=_torch.float, device=self.device)
        mean = _torch.tensor(self.mean, dtype=_torch.float, device=self.device)
        std = _torch.tensor(self.std, dtype=_torch.float, device=self.device)
        # tensor = tensor.clone().detach().requires_grad_(True)
        tensor = _torch.nn.Parameter(tensor)
        print("max ", tensor.max().item(), "\tmin ", tensor.min().item())
        optimizer = _torch.optim.Adam([tensor], lr=lr, weight_decay=L2)
        # Set the hook on the model
        for module in self.convolution_layers:
            module.low_memory = False
        layers = self.convolution_layers+self.dense_layers
        watcher = _common.LayerActivationWatcher(layers[layer])
        # loop until the layer's activation is maximised
        images = [input_image]
        f = save_frequency
        if verbose:
            print()
        try:
            for i in range(iterations):
                optimizer.zero_grad()
                old = tensor.clone().detach()
                self((tensor.clamp(0., 255.) - mean)/std)
                loss = -(watcher.activation[:, channel]).mean()
                loss.backward()
                optimizer.step()
                print("max ", tensor.max().item(), "\tmin ",
                      tensor.min().item())
                # save the intermediate image
                if (i % f == 0) or i == iterations-1:
                    # images.append(self._tensor_to_images(tensor,
                    #               mean=self.mean, std=self.std)[0])
                    images.append(_ml.Image(tensor=tensor.detach().cpu()[0]))
                    if verbose:
                        print("image saved")
                # compute the change
                change = _torch.mean((tensor - old)**2).item()
                change /= _torch.mean(old**2).item()
                if verbose:
                    print(f"\riteration {i}: change={change:.3g}\t\t\t")
        except KeyboardInterrupt:
            pass
        if verbose:
            print()
        return images

    def forward(self, tensor):
        """Don't use this"""
        for i, conv in enumerate(self.convolution_layers):
            tensor = conv(tensor)
        tensor = tensor.view(len(tensor), -1)
        for i, dense in enumerate(self.dense_layers):
            tensor = dense(tensor)
        return tensor

    def _to_tensor(self, data):
        """Converts a tuple (images, labels) into a tuple of tensors"""
        images, labels = data
        X = self._images_to_tensor(images, mean=self.mean, std=self.std)
        Y = self._labels_to_tensor(labels)
        return X, Y

    def _preprocess(self, training, validation):
        """preprocess the data (get image shape, normalization, ...)"""
        # batchify if needed
        if not isinstance(training, _types.FunctionType):
            return self._preprocess(lambda: (yield training),
                                    lambda: (yield validation))
        # treat by batches
        sum_, sum2, n = 0., 0., 0.
        shapes = []
        categories = set()
        # treat training data
        for i, (images, labels) in enumerate(training()):
            sum_, sum2, n = self._sums(images, sum_, sum2, n)
            categories = categories | set(labels)
            for image in images:
                shape = image.tensor.shape
                if shape not in shapes:
                    shapes.append(shape)
        n_training = i+1
        # treat validation data
        for i, (images, labels) in enumerate(validation()):
            sum_, sum2, n = self._sums(images, sum_, sum2, n)
            categories = categories | set(labels)
            for image in images:
                shape = image.tensor.shape
                if shape not in shapes:
                    shapes.append(shape)
        n_validation = i+1
        mean, std = self._mean_std(sum_, sum2, n)
        if len(shapes) > 1:
            raise ValueError(f"Expected a single image shape, found {shapes}")
        return mean, std, shapes.pop(), list(categories), \
            n_training, n_validation

    def _set_layers(self, *args):
        """Remove the previous layers and create new ones"""
        windows, channels, strides, pooling, dense, non_linear, \
            padded = args
        self.args = args
        # Deleting previous layers
        for i, _ in enumerate(self.convolution_layers):
            delattr(self, f"conv{i}")
        self.convolution_layers = []
        for i, _ in enumerate(self.dense_layers):
            delattr(self, f"dense{i}")
        self.dense_layers = []
        # Set convolution layers
        conv = _layers.ConvolutionLayer
        c, h, w = self.input_shape
        for i, c_out in enumerate(channels):
            layer = conv(c, c_out, non_linear=non_linear,
                         convolution_window=windows[i],
                         padding_size="auto" if padded else (0, 0, 0, 0),
                         stride=strides[i], pooling_window=pooling[i])
            self.convolution_layers.append(layer)
            setattr(self, f"conv{i}", layer)
            c, h, w = layer.shape_out((c, h, w))
            if c*h*w == 0:
                raise ValueError(f"image shape is {(c, h, w)} after conv{i}")
        # Set dense layers
        dens = _layers.DenseLayer
        n = c*h*w
        for i, n_out in enumerate(dense):
            layer = dens(n, n_out, non_linear=non_linear)
            self.dense_layers.append(layer)
            setattr(self, f"dense{i}", layer)
            n = layer.shape_out
        # out layer
        i = len(self.dense_layers)
        n_out = len(self.categories)
        layer = dens(n, n_out, non_linear="identity")
        self.dense_layers.append(layer)
        setattr(self, f"dense{i}", layer)
        # migrate to appropriate device
        self.to(self.device)
        # reset optimization
        self.reset()

    @property
    def shapes(self):
        """return the shape of the image after the successive layers"""
        shape = self.input_shape
        shapes = [shape]
        for conv in self.convolution_layers:
            shape = conv.shape_out(shape)
            shapes.append(shape)
        for dense in self.dense_layers:
            shapes.append(dense.linear.weight.shape[0])
        return shapes

    @property
    def dump(self):
        parameters = dict()
        parameters["args"] = self.args
        parameters["categories"] = self.categories
        parameters["mean"] = self.mean.tolist()
        parameters["std"] = self.std.tolist()
        parameters["input_shape"] = self.input_shape
        conv = [c.dump for c in self.convolution_layers]
        parameters["convolution_layers"] = conv
        dense = [d.dump for d in self.dense_layers]
        parameters["dense_layers"] = dense
        return {type(self).__name__: parameters, "version": _ml.__version__}

    @dump.setter
    def dump(self, other):
        if type(self).__name__ not in other.keys():
            raise ValueError(f"Expected dump from a '{type(self).__name__}'")
        other = other[type(self).__name__]
        self.categories = other["categories"]
        self.mean = _np.array(other["mean"])
        self.std = _np.array(other["std"])
        self.input_shape = other["input_shape"]
        self.args = other["args"]
        if self.args is not None:
            self._set_layers(*self.args)
        for i, d in enumerate(other["convolution_layers"]):
            self.convolution_layers[i].dump = d
        for i, d in enumerate(other["dense_layers"]):
            self.dense_layers[i].dump = d
        self.reset_optimizer()



import torch
import pandas as pd
import numpy as np
import torch.nn.functional as F
from typing import List, Union
from .layers import BatchNorm1d, Linear, FullyConnected
from .conversions import dataframe_to_tensor, categories_to_tensor, \
                         floats_to_tensor, tensor_to_categories
from .nn_decorators import nn_classifier


@nn_classifier
class DenseClassifier(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        pass

    def __init__(self, inputs: List[str], categories: List[str],
                 hidden_layers: List[int] = [10, 10, 10],
                 activation: str = "relu"):
        super().__init__()
        self.inputs = list(inputs)
        self.input_norm = BatchNorm1d(len(inputs))
        self.hidden_layers = FullyConnected(len(inputs),
                                            hidden_layers=hidden_layers,
                                            activation=activation)
        self.categories = list(categories)
        self.output = Linear(self.hidden_layers.out_features,
                             len(self.categories))

    def forward(self, x):
        x = self.input_norm(x)
        x = self.hidden_layers(x)
        return self.output(x)

    def data_to_tensor(self, X: pd.DataFrame,
                       Y: Union[None, List[str]],
                       weights: Union[None, List[float]] = None
                       ) -> tuple:
        x = dataframe_to_tensor(X, self.inputs, self.device)
        y = None if Y is None else categories_to_tensor(Y, self.categories,
                                                        self.device)
        w = None if weights is None else floats_to_tensor(weights, self.device)
        return x, y, w

    def tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        return tensor_to_categories(tensor, self.categories)

    def loss(self, y_pred: torch.Tensor, y_target: torch.Tensor,
             weights: Union[None, torch.Tensor]) -> torch.Tensor:
        if weights is None:
            return F.cross_entropy(y_pred, y_target, weight=self.class_weights)
        else:
            return F.nll_loss(F.log_softmax(y_pred) * weights, y_target,
                              weight=self.class_weights)

    @property
    def dump(self):
        pass



import types as _types
import torch as _torch
import machine_learning as _ml
import machine_learning.neural_networks.common as _common
import machine_learning.neural_networks.layers as _layers
import numpy as _np


class ImageTransformer(_common.NNtemplate):
    """
    A model that applies a transformation over an image.

    Attributes:
    -----------
    convolution_layers : list
        The list of ConvolutionLayer objects
    channels_in : int
        channels of the input image
    channels_out : int
        channels of the output image
    mean_x : np.ndarray
        The mean value for normalization of each input channel
    std_x : np.ndarray
        The standard deviation for normalization of each input channel
    mean_y : np.ndarray
        The mean value for normalization of each output channel
    std_y : np.ndarray
        The standard deviation for normalization of each output channel
    args : list or None
        list of arguments used to generate the model
    """

    def __init__(self):
        super().__init__()
        self.convolution_layers = []
        self.channels_in = 0
        self.channels_out = 0
        self.args = None

    def fit(self, images, masks, validation=None, validation_fraction=0.2,
            windows=[], channels=[], restart=True, non_linear="relu",
            l_max=None, low_memory=False, **kwargs):
        """
        Fit the model on the given data.

        Parameters:
        -----------
        images : list of Image objects
            a list of machine_learning.Images objects
        masks : list of Image objects
            a list of arrays of shape (channels, height, width)
        validation : tuple of (images, masks) or None
            data used for validation
        validation_fraction : float
            if validation is None, training data is splited
        windows : list of tuple of int
            the convolution windows (height, width)
        channels : list of int
            channels number at output of each convolution
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        l_max : int or None
            maximum observations in a batch
        low_memory : bool
            If True, save memory at the cost of computation speed.
            Has effect only for deep networks.
        Returns:
        --------
        training, validation : tuples of (images, masks)

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        training = (images, masks)
        training, validation = self._validation_split(training, validation,
                                                      validation_fraction)
        res = self._preprocess(training, validation)
        self.mean_x, self.std_x, self.mean_y, self.std_y, self.channels_in, \
            self.channels_out, _, _ = res
        # create layers
        if restart:
            self._set_layers(windows, channels, non_linear, low_memory)
        # load data in memory
        train = self._to_tensor(training)
        val = self._to_tensor(validation)
        train, n_training = self._wrap_batch(train, l_max)
        val, n_validation = self._wrap_batch(val, l_max)
        # fitting
        self.train_loop(train, val, n_training, n_validation, **kwargs)
        # returning training and validation data
        return training, validation

    def fit_batch(self, training, validation, windows=[], channels=[],
                  restart=True, non_linear="relu", low_memory=False, **kwargs):
        """
        Fit the model on the given batches.

        Parameters:
        -----------
        training : generator function
            function that yields (images, labels)
        validation : generator function
            function that yields (images, labels)
        windows : list of tuple of int
            the convolution windows (height, width)
        channels : list of int
            channels number at output of each convolution
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        low_memory : bool
            If True, save memory at the cost of computation speed.
            Has effect only for deep networks.

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        res = self._preprocess(training, validation)
        self.mean_x, self.std_x, self.mean_y, self.std_y, self.channels_in, \
            self.channels_out, n_training, n_validation = res
        # create layers
        if restart:
            self._set_layers(windows, channels, non_linear, low_memory)
        # fitting
        self.train_loop(self._wrap(training), self._wrap(validation),
                        n_training, n_validation, **kwargs)

    def predict(self, images):
        """
        Make a prediction using the trained model.

        Parameters:
        -----------
        images : Image or list of Images

        Returns:
        --------
        labels : list
        """
        scalar = False
        if isinstance(images, _ml.Image):
            images = [images]
            scalar = True
        tensor = self._images_to_tensor(images, self.mean_x, self.std_x)
        with _torch.no_grad():
            output = self(tensor)
        masks = self._tensor_to_images(output, self.mean_y, self.std_y)
        if scalar:
            masks = masks[0]
        return masks

    def forward(self, tensor):
        """Don't use this"""
        for i, conv in enumerate(self.convolution_layers):
            tensor = conv(tensor)
        return tensor

    def _to_tensor(self, data):
        """Converts a tuple (images, masks) into a tuple of tensors"""
        images, masks = data
        X = self._images_to_tensor(images, mean=self.mean_x, std=self.std_x)
        Y = self._images_to_tensor(masks, mean=self.mean_y, std=self.std_y)
        return X, Y

    def _preprocess(self, training, validation):
        """preprocess the data (get image shape, normalization, ...)"""
        # batchify if needed
        if not isinstance(training, _types.FunctionType):
            return self._preprocess(lambda: (yield training),
                                    lambda: (yield validation))
        # treat by batches
        sum_x, sum2_x, n_x = 0., 0., 0.
        sum_y, sum2_y, n_y = 0., 0., 0.
        channels_in = set()
        channels_out = set()
        # treat training data
        for i, (images, masks) in enumerate(training()):
            sum_x, sum2_x, n_x = self._sums(images, sum_x, sum2_x, n_x)
            sum_y, sum2_y, n_y = self._sums(masks, sum_y, sum2_y, n_y)
            for image in images:
                channels_in = channels_in | {image.tensor.shape[0]}
            for mask in masks:
                channels_out = channels_out | {mask.tensor.shape[0]}
        n_training = i+1
        # treat validation data
        for i, (images, masks) in enumerate(validation()):
            sum_x, sum2_x, n_x = self._sums(images, sum_x, sum2_x, n_x)
            sum_y, sum2_y, n_y = self._sums(masks, sum_y, sum2_y, n_y)
            for image in images:
                channels_in = channels_in | {image.tensor.shape[0]}
            for mask in masks:
                channels_out = channels_out | {mask.tensor.shape[0]}
        n_validation = i+1
        # returns values
        if len(channels_in) > 1:
            raise ValueError("found different put channels count: "
                             f"{channels_in}")
        if len(channels_out) > 1:
            raise ValueError("found different output channels count: "
                             f"{channels_out}")
        mean_x, std_x = self._mean_std(sum_x, sum2_x, n_x)
        mean_y, std_y = self._mean_std(sum_y, sum2_y, n_y)
        channels_in, channels_out = channels_in.pop(), channels_out.pop()
        return (mean_x, std_x, mean_y, std_y, channels_in, channels_out,
                n_training, n_validation)

    def _set_layers(self, *args):
        """Remove the previous layers and create new ones"""
        windows, channels, non_linear, low_memory = args
        self.args = args
        # Deleting previous layers
        for i, _ in enumerate(self.convolution_layers):
            delattr(self, f"conv{i}")
        self.convolution_layers = []
        # Set convolution layers
        conv = _layers.ConvolutionLayer
        c_in = self.channels_in
        for i, c_out in enumerate(channels):
            window = windows[i]
            layer = conv(c_in, c_out, non_linear=non_linear,
                         convolution_window=window, padding_size="auto",
                         low_memory=low_memory,
                         stride=(1, 1), pooling_window=(1, 1))
            self.convolution_layers.append(layer)
            setattr(self, f"conv{i}", layer)
            c_in = c_out
        # out layer
        c_out = self.channels_out
        i = len(self.convolution_layers)
        layer = conv(c_in, c_out, non_linear="identity",
                     convolution_window=(1, 1), padding_size="auto",
                     stride=(1, 1), pooling_window=(1, 1))
        self.convolution_layers.append(layer)
        setattr(self, f"conv{i}", layer)
        # migrate to appropriate device
        self.to(self.device)
        # reset optimization
        self.reset()

    @property
    def dump(self):
        parameters = dict()
        parameters["channels_in"] = self.channels_in
        parameters["channels_out"] = self.channels_out
        parameters["args"] = self.args
        parameters["mean_x"] = self.mean_x.tolist()
        parameters["std_x"] = self.std_x.tolist()
        parameters["mean_y"] = self.mean_y.tolist()
        parameters["std_y"] = self.std_y.tolist()
        conv = [conv.dump for conv in self.convolution_layers]
        parameters["convolution_layers"] = conv
        return {type(self).__name__: parameters, "version": _ml.__version__}

    @dump.setter
    def dump(self, other):
        if type(self).__name__ not in other.keys():
            raise ValueError(f"Expected dump from a '{type(self).__name__}'")
        other = other[type(self).__name__]
        self.channels_in = other["channels_in"]
        self.channels_out = other["channels_out"]
        self.mean_x = _np.array(other["mean_x"])
        self.std_x = _np.array(other["std_x"])
        self.mean_y = _np.array(other["mean_y"])
        self.std_y = _np.array(other["std_y"])
        self.args = other["args"]
        if self.args is not None:
            self._set_layers(*self.args)
        for i, d in enumerate(other["convolution_layers"]):
            self.convolution_layers[i].dump = d
        self.reset_optimizer()



from .._library_loader import _lib
from ..tensor import Tensor as _Tensor
import ctypes as _c
import numpy as _np
import matplotlib.pyplot as _plt
import os as _os

# Typing of the Tensor functions
_lib.new_Image_from_path.restype = _c.c_void_p
_lib.new_Image_from_path.argtypes = [_c.c_char_p]

_lib.new_Image_from_data.restype = _c.c_void_p
_lib.new_Image_from_data.argtypes = [_c.c_uint8, _c.c_uint64, _c.c_uint64,
                                     _c.c_uint64]

_lib.Image_height.restype = _c.c_uint64
_lib.Image_height.argtypes = [_c.c_void_p]

_lib.Image_width.restype = _c.c_uint64
_lib.Image_width.argtypes = [_c.c_void_p]

_lib.Image_channels.restype = _c.c_uint64
_lib.Image_channels.argtypes = [_c.c_void_p]

_lib.Image_as_grayscale.restype = _c.c_void_p
_lib.Image_as_grayscale.argtypes = [_c.c_void_p]

_lib.Image_as_RGB.restype = _c.c_void_p
_lib.Image_as_RGB.argtypes = [_c.c_void_p]

_lib.Image_as_RGBA.restype = _c.c_void_p
_lib.Image_as_RGBA.argtypes = [_c.c_void_p]

_lib.Image_resized.restype = _c.c_void_p
_lib.Image_resized.argtypes = [_c.c_void_p, _c.c_uint64, _c.c_uint64]

_lib.Image_save.restype = None
_lib.Image_save.argtypes = [_c.c_void_p]

_lib.Image_copy.restype = None
_lib.Image_copy.argtypes = [_c.c_void_p]

_lib.Image_as_Tensor.restype = _c.c_void_p
_lib.Image_as_Tensor.argtypes = [_c.c_void_p]

_lib.del_Image.restype = None
_lib.del_Image.argtypes = [_c.c_void_p]


# Defining the class
class Image:

    def __init__(self, data):
        if isinstance(data, _c.c_void_p):
            self.pointer = data
        elif isinstance(data, str):
            assert _os.path.isfile(data)
            self.pointer = _lib.new_Image_from_path(data.encode("utf-8"))
        elif hasattr(data, "__iter__"):
            data = _np.array(data, dypes=_np.uint8)
            if len(data.shape) == 2:
                data = data[:, :, None]
            assert len(data.shape) == 3
            height, width, channels = data.shape
            input = data.ctypes.as_type(_c.c_uint8)
            self.pointer = _lib.new_Image_from_data(input, height, width,
                                                    channels)
        else:
            self.pointer = None
            raise TypeError(f"Unexpected type '{type(data)}'"
                            "in Image's constructor")

    def __del__(self):
        _lib.del_Image(self.pointer)

    def draw(self, axis=None):
        if axis is None:
            _, axis = _plt.subplots()
        axis.imshow(self.numpy, cmap="gray")
        axis.set_xticks([])
        axis.set_yticks([])

    def as_grayscale(self):
        return Image(_c.c_void_p(_lib.Image_as_grayscale(self.pointer)))

    def as_RGB(self):
        return Image(_c.c_void_p(_lib.Image_as_RGB(self.pointer)))

    def as_RGBA(self):
        return Image(_c.c_void_p(_lib.Image_as_RGBA(self.pointer)))

    def resized(self, height: int, width: int):
        return Image(_c.c_void_p(_lib.Image_resized(self.pointer,
                                                    height, width)))

    def save(self, path: str):
        _lib.Image_save(path.encode("utf-8"))

    @property
    def tensor(self):
        return _Tensor(_c.c_void_p(_lib.Image_as_Tensor(self.pointer)))

    @property
    def numpy(self):
        arr = self.tensor.numpy
        if arr.shape[0] == 1:
            return arr.reshape(arr.shape[1:]).astype(_np.uint8)
        else:
            return _np.moveaxis(self.tensor.numpy, 0, -1).astype(_np.uint8)



import torch
from typing import Tuple, Union
from .convolution import Conv2d
from . import pooling as poo
from .activation import Activation
from .padding import ConstantPad2d
from.batch_norm import BatchNorm2d


class Downsampling(torch.nn.Module):
    """
    A downsampling layer is a sequence of:
    - a 'ConstantPad2d'
    - a 'Conv2d', layers
    - a pooling layer 'MaxPool2d' or 'AvgPool2d'
    - an 'Activation' layer
    - a 'BatchNorm2d' layer

    Attributes
    ----------
    padding : ConstantPad2d or None
        The padding layer
    convolution : Conv2d
        The convolution layers
    pooling : MaxPool2d or AvgPool2d or None
        The pooling layer
    activation : Activation
        The activation layer
    batch_norm : BatchNorm2d
        The batch normalization layer
    concatenate : bool
        Whether the input data should be resized
        and concatenated to the output
    """

    @classmethod
    def from_dump(cls, dump: dict) -> 'Downsampling':
        """returns a 'Downsampling' object from a dump"""
        assert dump["type"] == cls.__name__
        obj = cls(1, 1, (1, 1), (1, 1), None,
                  activation=dump["activation"]["function name"],
                  concatenate=dump["concatenate"])
        if dump["padding"] is not None:
            obj.padding = ConstantPad2d.from_dump(dump["padding"])
        obj.convolutions = Conv2d.from_dump(dump["convolution"])
        if dump["pooling"] is not None:
            pooling_type = getattr(poo, dump["pooling"]["type"])
            obj.pooling = pooling_type(dump["pooling"])
        obj.activation = Activation.from_dump(dump["activation"])
        obj.batch_norm = BatchNorm2d.from_dump(dump["batch norm"])
        return obj

    def __init__(self, in_channels: int,
                 out_channels: int,
                 conv_size: Tuple[int, int],
                 conv_stride: Tuple[int, int],
                 pooling_size: Union[None, Tuple[int, int]],
                 activation: str = "relu",
                 padded: bool = True,
                 pooling_kind: str = "Max",
                 concatenate: bool = False):
        """
        Parameters
        ----------
        in_channels : int
            The input channels
        out_channels : int
            The output channels of the successive convolutions
        conv_size : tuple of int
            The (height, width) of convolution window
        conv_stride : tuple of int
            The (height, width) stride of the convolution
        pooling_size : tuple of int or None
            The shape of the pooling window
            If None, no pooling is performed
        activation : str
            The name of the activation function
        padded : bool
            Whether padding should be applied before each convolution.
            If True, padding size is set to keep (height, width) constant
            after each convolution.
        pooling_kind : str
            The type of pooling layer.
            Must be one of ("Max", "Avg")
        concatenate : bool
            Whether the input data should be resized
            and concatenated to the output
        """
        super().__init__()
        if padded:
            h, w = conv_size
            dh, dw = h//2, w//2
            self.padding = ConstantPad2d((dw, dw, dh, dh), 0.)
        else:
            self.padding = None
        self.convolution = Conv2d(in_channels, out_channels,
                                  conv_size, conv_stride)
        if pooling_size is None:
            self.pooling = None
        else:
            self.pooling = getattr(poo, pooling_kind+"Pool2d")(pooling_size)
        self.activation = Activation(activation)
        self.batch_norm = BatchNorm2d(out_channels)
        self.concatenate = concatenate

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        X = input
        if self.padding is not None:
            X = self.padding(X)
        X = self.convolution(X)
        if self.pooling is not None:
            X = self.pooling(X)
        X = self.batch_norm(self.activation(X))
        if self.concatenate:
            Y = input
            # Triming the sides of the input
            if self.padding is None:
                h, w = self.convolution.kernel_size
                dh, dw = h//2, w//2
                Y = Y[:, :, dh:-dh, dw:-dw]
            # Average pool the input to resize to same size as the output
            h, w = self.convolution.kernel_size
            if self.pooling is not None:
                hp, wp = self.pooling.kernel_size
                h, w = h*hp, w*wp
            Y = torch.nn.functional.avg_pool2d(Y, (h, w))
            # Concatenate over the X
            X = torch.cat((Y, X), dim=1)
        # Returns the resulting tensor
        return X

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "padding": None if self.pading is None else self.padding.dump,
                "convolution": self.convolution.dump,
                "pooling": None if self.pooling is None else self.pooling.dump,
                "activation": self.activation.dump,
                "batch norm": self.batch_norm.dump,
                "concatenate": self.concatenate}



import torch
from typing import Tuple
from .padding import ConstantPad1d, ConstantPad2d, load_padding
from .convolution import Conv1d, Conv2d, load_convolution
from .activation import Activation
from .batch_norm import BatchNorm1d, BatchNorm2d, load_batchnorm
from .pooling import AvgPool1d, AvgPool2d, load_pooling


def load_activatedconv(dump: dict):
    """returns the ConvStage object loaded from a dump"""
    return globals()[dump["type"]].from_dump(dump)


class _ConvStage(torch.nn.Module):
    """A template for activated convolutions"""

    @classmethod
    def from_dump(cls, dump: dict) -> '_ConvStage':
        """returns a '_ConvStageolution' layer from a dump"""
        assert dump["type"] == cls.__name__
        obj = cls(1)
        obj.padded = dump["padded"]
        obj.stacked = dump["stacked"]
        obj.padding = load_padding(dump["padding"])
        obj.convolution = load_convolution(dump["convolution"])
        obj.activation = Activation.from_dump(dump["activation"])
        obj.batch_norm = load_batchnorm(dump["batch norm"])
        obj.downsampler = load_pooling(dump["downsampler"])
        return obj

    def __init__(self):
        super().__init__()

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        if self.padded:
            X = self.padding(X)
        Y = self.batch_norm(self.activation(self.convolution(X)))
        if self.stacked:
            resized = self.downsampler(X)
            Y = torch.cat([resized, Y], dim=1)
        return Y

    def shape_out(self, shape_in: list) -> list:
        shape_out = self.padding.shape_out(shape_in)
        shape_out = self.convolution.shape_out(shape_out)
        if self.stacked:
            shape_out[0] += self.out_channels
        return shape_out

    def shape_in(self, shape_out: list) -> list:
        if self.stacked:
            shape_out[0] -= self.out_channels
        shape_in = self.convolution.shape_in(shape_out)
        shape_in = self.padding.shape_in(shape_in)
        return shape_in

    @property
    def in_channels(self):
        return self.convolution.in_channels

    @property
    def out_channels(self):
        c = self.convolution.out_channels
        if self.stacked:
            c += self.convolution.in_channels
        return c

    @property
    def window(self):
        return self.convolution.kernel_size

    @property
    def stride(self):
        return self.convolution.stride

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "padded": self.padded,
                "stacked": self.stacked,
                "padding": self.padding.dump,
                "convolution": self.convolution.dump,
                "activation": self.activation.dump,
                "batch norm": self.batch_norm.dump,
                "downsampler": self.downsampler.dump}


class ConvStage1d(_ConvStage):

    def __init__(self, in_channels: int,
                 channels: int = 8,
                 window: int = 3,
                 stride: int = 1,
                 activation: str = "relu",
                 padded: bool = True,
                 with_bias: bool = True,
                 stacked: bool = False):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        channels : int
            The number of output channels
        window : int
            The size of the convolved kernel
        stride : int
            The stride used for the convolution
        padded : bool
            Whether to pad so that the output has the same size as input
        with_bias : bool
            Whether to have a bias in the linear equation:
            y = weight*x + bias
        stacked: bool
            If True the input is downsampled and concatenated to the output
        """
        super().__init__()
        self.padded = padded
        self.stacked = stacked
        left = (window-1)//2
        right = window-1 - left
        pad = (left, right) if padded else (0, 0)
        self.padding = ConstantPad1d(pad, value=0.)
        self.convolution = Conv1d(in_channels, channels, window, stride,
                                  with_bias=with_bias)
        self.activation = Activation(activation)
        self.batch_norm = BatchNorm1d(channels)
        self.downsampler = AvgPool1d(window, stride)


class ConvStage2d(_ConvStage):

    def __init__(self, in_channels: int,
                 channels: int = 8,
                 window: Tuple[int, int] = (3, 3),
                 stride: Tuple[int, int] = (1, 1),
                 activation: str = "relu",
                 padded: bool = True,
                 with_bias: bool = True,
                 stacked: bool = False):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        channels : int
            The number of output channels
        window : tuple of int
            The (height, width) size of the convolved kernel
        stride : tuple of int
            The (height, width) stride used for the convolution
        padded : bool
            Whether to pad so that the output has the same size as input
        with_bias : bool
            Whether to have a bias in the linear equation:
            y = weight*x + bias
        stacked: bool
            If True the input is downsampled and concatenated to the output
        """
        super().__init__()
        self.padded = padded
        self.stacked = stacked
        left, top = ((w-1)//2 for w in window)
        right, bottom = (w-1 - first for w, first in zip(window, [left, top]))
        pad = (left, right, top, bottom) if padded else (0, 0, 0, 0)
        self.padding = ConstantPad2d(pad, value=0.)
        self.convolution = Conv2d(in_channels, channels, window, stride,
                                  with_bias=with_bias)
        self.activation = Activation(activation)
        self.batch_norm = BatchNorm2d(channels)
        self.downsampler = AvgPool2d(window, stride)



import torch
import torch.nn.functional as F
from typing import Tuple, List


def load_pooling(dump):
    """returns the Pooling object loaded from a dump"""
    return globals()[dump["type"]].from_dump(dump)


class _Pooling:
    """A template for pooling layers"""

    @classmethod
    def from_dump(cls, dump: dict) -> '_Pooling':
        """returns a '_Pooling' layer from a dump"""
        assert dump["type"] == cls.__name__
        return cls(dump["kernel size"], stride=dump["stride"])

    @property
    def dump(self) -> dict:
        ks = self.kernel_size
        st = self.stride
        return {"type": type(self).__name__,
                "kernel size": list(ks) if hasattr(ks, "__iter__") else ks,
                "stride": list(st) if hasattr(st, "__iter__") else st}

    def shape_out(self, shape_in: list) -> list:
        c = shape_in[0]
        dims = shape_in[1:]
        dims = [int((d - k)/s + 1) for d, k, s in
                zip(dims, self.kernel_size, self.stride)]
        return [c] + dims

    def shape_in(self, shape_out: list) -> list:
        c = shape_out[0]
        dims = shape_out[1:]
        dims = [(d-1)*s + k for d, k, s in
                zip(dims, self.kernel_size, self.stride)]
        return [c] + dims


class MaxPool1d(torch.nn.MaxPool1d, _Pooling):
    """A wrapper around torch.nn.MaxPool1d"""

    def __init__(self, kernel_size: int, stride: int = 1):
        """
        Parameters
        ----------
        kernel_size : int
            The size of the pooling window
        stride : int
            The stride of the pooling window
        """
        torch.nn.MaxPool1d.__init__(self, kernel_size, stride)


class MaxPool2d(torch.nn.MaxPool2d, _Pooling):
    """A wrapper around torch.nn.MaxPool2d"""

    def __init__(self, kernel_size: Tuple[int, int], stride: int = 1):
        """
        Parameters
        ----------
        kernel_size : tuple of int
            The (height, width) size of the pooling window
        stride : tuple of int
            The (height, width) stride of the pooling window
        """
        torch.nn.MaxPool1d.__init__(self, kernel_size, stride)


class AvgPool1d(torch.nn.AvgPool1d, _Pooling):
    """A wrapper around torch.nn.AvgPool1d"""

    def __init__(self, kernel_size: int, stride: int = 1):
        """
        Parameters
        ----------
        kernel_size : int
            The size of the pooling window
        stride : int
            The stride of the pooling window
        """
        torch.nn.AvgPool1d.__init__(self, kernel_size, stride)


class AvgPool2d(torch.nn.AvgPool2d, _Pooling):
    """A wrapper around torch.nn.AvgPool2d"""

    def __init__(self, kernel_size: Tuple[int, int], stride: int = 1):
        """
        Parameters
        ----------
        kernel_size : tuple of int
            The (height, width) size of the pooling window
        stride : tuple of int
            The (height, width) stride of the pooling window
        """
        torch.nn.AvgPool2d.__init__(self, kernel_size, stride)


class _OverallPool(torch.nn.Module):
    """
    Over all pooling transforms a tensor of shape (N, C, ...)
    to a tensor of shape (N, C)
    """

    @classmethod
    def from_dump(cls, dump: dict) -> 'OverallPool1d':
        assert dump["type"] == cls.__name__
        return cls(dump["pooling type"])

    def __init__(self, pooling_type: str, dimension: str):
        """
        Parameters
        ----------
        pooling type : one of {"Max", "Avg"}
            the type of pooling to perform
        dimension : one of {"1d", "2d"}
            the dimension of the tensor to treat
        """
        super().__init__()
        self.pooling_type = pooling_type
        self.func = getattr(F, pooling_type.lower()+"_pool"+dimension)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        return self.func(X, X.shape[2:]).view(X.shape[:2])

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "pooling type": self.pooling_type}

    def shape_in(self, shape_out: int) -> List[int]:
        return [shape_out, 1, 1]

    def shape_out(self, shape_in: List[int]) -> int:
        return shape_in[0]


class OverallPool1d(_OverallPool):

    def __init__(self, pooling_type: str):
        """
        Parameters
        ----------
        pooling type : one of {"Max", "Avg"}
            the type of pooling to perform
        """
        super().__init__(pooling_type, "1d")


class OverallPool2d(_OverallPool):

    def __init__(self, pooling_type: str):
        """
        Parameters
        ----------
        pooling type : one of {"Max", "Avg"}
            the type of pooling to perform
        """
        super().__init__(pooling_type, "2d")




import types as _types
import torch as _torch
import machine_learning as _ml
import machine_learning._templates as _templates
import machine_learning.neural_networks.common as _common
import machine_learning.neural_networks.layers as _layers
import numpy as _np


class ImageSegmenter(_common.NNtemplate, _templates.ModelTemplate):
    """
    A model that returns a segmented grayscale image.
    Each gray level of the output corresponds to a class.
    (for example output can be a binarized 0 and 255 gray levels image)

    Attributes:
    -----------
    categories : list of uint8
        gray levels the output image can be segmented into
    convolution_layers : list
        The list of ConvolutionLayer objects
    channels_in : int
        channels of the input image
    mean : np.ndarray
        The mean value for normalization of each input channel
    std : np.ndarray
        The standard deviation for normalization of each input channel
    args : list or None
        list of arguments used to generate the model
    """

    def __init__(self):
        super().__init__()
        self.convolution_layers = []
        self.upsampling_layers = []
        self.out_layer = None
        self.channels_in = 0
        self.loss = _torch.nn.functional.cross_entropy
        self.args = None

    def fit(self, images, masks, validation=None, validation_fraction=0.2,
            down_windows=[], down_channels=[], pooling_windows=[],
            up_windows=[], up_channels=[], restart=True, non_linear="relu",
            l_max=None, **kwargs):
        """
        Fit the model on the given data.

        Parameters:
        -----------
        images : list of Image objects
            a list of machine_learning.Images objects
        masks : list of Image objects
            a list of arrays of shape (channels, height, width)
        validation : tuple of (images, masks) or None
            data used for validation
        validation_fraction : float
            if validation is None, training data is splited
        windows : list of tuple of int
            the convolution windows (height, width)
        channels : list of int
            channels number at output of each convolution
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        l_max : int or None
            maximum observations in a batch
        low_memory : bool
            If True, save memory at the cost of computation speed.
            Has effect only for deep networks.
        Returns:
        --------
        training, validation : tuples of (images, masks)

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        training = (images, masks)
        training, validation = self._validation_split(training, validation,
                                                      validation_fraction)
        res = self._preprocess(training, validation)
        (self.mean, self.std, self.channels_in, self.categories,
            frequ, _, _) = res
        # redefine loss with correct weights
        weights = _torch.Tensor([1/f for f in frequ])
        self.loss = _torch.nn.CrossEntropyLoss(weight=weights)
        # create layers
        if restart:
            self._set_layers(down_windows, down_channels, pooling_windows,
                             up_windows, up_channels, non_linear)
        # load data in memory
        train = self._to_tensor(training)
        val = self._to_tensor(validation)
        train, n_training = self._wrap_batch(train, l_max)
        val, n_validation = self._wrap_batch(val, l_max)
        # fitting
        self.train_loop(train, val, n_training, n_validation, **kwargs)
        # returning training and validation data
        return training, validation

    def fit_batch(self, training, validation, down_windows=[],
                  down_channels=[], pooling_windows=[], up_windows=[],
                  up_channels=[], restart=True, non_linear="relu", **kwargs):
        """
        Fit the model on the given batches.

        Parameters:
        -----------
        training : generator function
            function that yields (images, labels)
        validation : generator function
            function that yields (images, labels)
        windows : list of tuple of int
            the convolution windows (height, width)
        channels : list of int
            channels number at output of each convolution
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        low_memory : bool
            If True, save memory at the cost of computation speed.
            Has effect only for deep networks.

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        res = self._preprocess(training, validation)
        (self.mean, self.std, self.channels_in, self.categories,
            frequ, n_training, n_validation) = res
        # redefine loss with correct weights
        weights = _torch.Tensor([1/f for f in frequ])
        self.loss = _torch.nn.CrossEntropyLoss(weight=weights)
        # create layers
        if restart:
            self._set_layers(down_windows, down_channels, pooling_windows,
                             up_windows, up_channels, non_linear)
        # fitting
        self.train_loop(self._wrap(training), self._wrap(validation),
                        n_training, n_validation, **kwargs)

    def predict(self, images):
        """
        Make a prediction using the trained model.

        Parameters:
        -----------
        images : Image or list of Images

        Returns:
        --------
        labels : list
        """
        scalar = False
        if isinstance(images, _ml.Image):
            images = [images]
            scalar = True
        tensor = self._images_to_tensor(images, self.mean, self.std)
        with _torch.no_grad():
            output = self(tensor)
        masks = self._tensor_to_segmented_images(output)
        if scalar:
            masks = masks[0]
        return masks

    def forward(self, tensor):
        """Don't use this"""
        downsampling = []
        for conv in self.convolution_layers:
            downsampling.append(tensor)
            tensor = conv(tensor)
        downsampling.reverse()
        for i, deconv in enumerate(self.upsampling_layers):
            tensor = deconv(tensor, downsampling[i])
        return self.out_layer(tensor)

    def _to_tensor(self, data):
        """Converts a tuple (images, masks) into a tuple of tensors"""
        images, masks = data
        X = self._images_to_tensor(images, mean=self.mean, std=self.std)
        Y = self._segmented_images_to_tensor(masks)
        return X, Y

    def _preprocess(self, training, validation):
        """preprocess the data (get image shape, normalization, ...)"""
        # batchify if needed
        if not isinstance(training, _types.FunctionType):
            return self._preprocess(lambda: (yield training),
                                    lambda: (yield validation))
        # treat by batches
        sum_, sum2, n = 0., 0., 0.
        channels_in = set()
        categories = []
        frequencies = []
        # treat training data
        for i, (images, masks) in enumerate(training()):
            sum_, sum2, n = self._sums(images, sum_, sum2, n)
            for image in images:
                channels_in = channels_in | {image.tensor.shape[0]}
            masks = _np.concatenate([mask.data for mask in masks])
            masks = masks.reshape((-1,) + masks.shape[2:])
            uniques, counts = _np.unique(masks, axis=0, return_counts=True)
            for u, c in zip(uniques, counts):
                u = u.tolist()
                try:
                    index = categories.index(u)
                except ValueError:
                    index = len(categories)
                    categories.append(u)
                    frequencies.append(c)
                finally:
                    frequencies[index] += c
        n_training = i+1
        # treat validation data
        for i, (images, masks) in enumerate(validation()):
            sum_, sum2, n = self._sums(images, sum_, sum2, n)
            for image in images:
                channels_in = channels_in | {image.tensor.shape[0]}
            masks = _np.concatenate([mask.data for mask in masks])
            masks = masks.reshape((-1,) + masks.shape[2:])
            uniques, counts = _np.unique(masks, axis=0, return_counts=True)
            for u, c in zip(uniques, counts):
                u = u.tolist()
                try:
                    index = categories.index(u)
                except ValueError:
                    index = len(categories)
                    categories.append(u)
                    frequencies.append(c)
                finally:
                    frequencies[index] += c
        n_validation = i+1
        # treating categories
        s = sum(frequencies)
        frequencies = [f/s for f in frequencies]
        # returns values
        if len(channels_in) > 1:
            raise ValueError("found different put channels count: "
                             f"{channels_in}")
        mean, std = self._mean_std(sum_, sum2, n)
        return (mean, std, channels_in.pop(), categories, frequencies,
                n_training, n_validation)

    def _set_layers(self, *args):
        """Remove the previous layers and create new ones"""
        (down_windows, down_channels, pooling_windows, up_windows, up_channels,
            non_linear) = args
        self.args = args
        # Deleting previous layers
        for i, _ in enumerate(self.convolution_layers):
            delattr(self, f"conv{i}")
        self.convolution_layers = []
        for i, _ in enumerate(self.upsampling_layers):
            delattr(self, f"upsample{i}")
        self.upsampling_layers = []
        # Setting alias
        conv = _layers.ConvolutionLayer
        up = _layers.UpsamplingLayer
        # Set convolution layers
        c_in = self.channels_in
        for i, c_out in enumerate(down_channels):
            window = down_windows[i]
            pooling = pooling_windows[i]
            layer = conv(c_in, c_out, non_linear=non_linear,
                         convolution_window=window, padding_size="auto",
                         stride=(1, 1), pooling_window=pooling)
            self.convolution_layers.append(layer)
            setattr(self, f"conv{i}", layer)
            c_in = c_out
        # Set upsampling_layer
        down_channels.reverse()
        down_channels = down_channels[1:] + [self.channels_in]
        for i, c_out in enumerate(up_channels):
            window = up_windows[i]
            layer = up(c_in+down_channels[i], c_out, non_linear=non_linear,
                       convolution_window=window)
            self.upsampling_layers.append(layer)
            setattr(self, f"upsample{len(self.upsampling_layers)}", layer)
            c_in = c_out
        # out layer
        c_out = len(self.categories)
        layer = conv(c_in, c_out, non_linear="identity",
                     convolution_window=(1, 1), padding_size="auto",
                     stride=(1, 1), pooling_window=(1, 1))
        self.out_layer = layer
        # migrate to appropriate device
        self.to(self.device)
        # reset optimization
        self.reset()

    @property
    def dump(self):
        parameters = dict()
        parameters["categories"] = self.categories
        parameters["mean"] = self.mean.tolist()
        parameters["std"] = self.std.tolist()
        parameters["args"] = self.args
        parameters["channels_in"] = self.channels_in
        conv = [conv.dump for conv in self.convolution_layers]
        parameters["convolution_layers"] = conv
        up = [up.dump for up in self.upsampling_layers]
        parameters["upsampling_layers"] = up
        parameters["out_layer"] = self.out_layer.dump
        return {type(self).__name__: parameters, "version": _ml.__version__}

    @dump.setter
    def dump(self, other):
        if type(self).__name__ not in other.keys():
            raise ValueError(f"Expected dump from a '{type(self).__name__}'")
        other = other[type(self).__name__]
        self.categories = other["categories"]
        self.mean = _np.array(other["mean"])
        self.std = _np.array(other["std"])
        self.args = other["args"]
        self.channels_in = other["channels_in"]
        if self.args is not None:
            self._set_layers(*self.args)
        for i, d in enumerate(other["convolution_layers"]):
            self.convolution_layers[i].dump = d
        for i, d in enumerate(other["upsampling_layers"]):
            self.upsampling_layers[i].dump = d
        self.out_layer.dump = other["out_layer"]
        self.reset_optimizer()



import machine_learning.neural_networks.common as _common
import machine_learning._templates as _templates
import torch as _torch
import types as _types
import machine_learning as _ml
import machine_learning.neural_networks.layers as _layers
import numpy as _np


class Regressor(_common.NNtemplate, _templates.RegressorTemplate):
    """
    A model that predict scalar values from vectors of observations.

    Attributes:
    -----------
    x : list of str
        name of x columns
    mean_x : np.ndarray
        The mean value for normalization of input
    std_x : np.ndarray
        The standard deviation for normalization of input
    mean_y : np.ndarray
        The mean value for normalization of output
    std_y : np.ndarray
        The standard deviation for normalization of output
    dense_layers : list
        list of the dense layers of the model
    args : list or None
        list of arguments used to generate the model
    """

    def __init__(self):
        super().__init__()
        self.dense_layers = []
        self.x = []
        self.args = None

    def fit(self, x, y, layers=(32, 16, 8), validation=None,
            validation_fraction=0.2, restart=True, non_linear="relu",
            l_max=None, **kwargs):
        """
        Fit the model on the given data.

        Parameters:
        -----------
        x : pandas.DataFrame
            real valued observations
        y : iterable
            scalar predictions
        layers : list of int
            The number of neurons in each dense hidden layer
        validation : pandas.DataFrame or None
            data used for validation
        validation_fraction : float
            if validation is None, training data is splited
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        l_max : int or None
            maximum observations in a batch

        Returns:
        --------
        training, validation : pandas.DataFrame

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        training = x, y
        # preprocess
        training, validation = self._validation_split(training, validation,
                                                      validation_fraction)
        res = self._preprocess(training, validation)
        self.mean_x, self.std_x, self.mean_y, self.std_y, self.x, _, _ = res
        # create layers
        if restart:
            self._set_layers(layers, non_linear)
        # load data in memory
        train = self._to_tensor(training)
        val = self._to_tensor(validation)
        train, n_training = self._wrap_batch(train, l_max)
        val, n_validation = self._wrap_batch(val, l_max)
        # fitting
        self.train_loop(train, val, n_training, n_validation, **kwargs)
        # returning training and validation data
        return training, validation

    def fit_batch(self, training, validation, x, y, layers,
                  restart=True, non_linear="relu", **kwargs):
        """
        Fit the model on the given batches.

        Parameters:
        -----------
        training : generator function
            function that yields a pandas dataframe
        validation : generator function
            function that yields a pandas dataframe
        x : list of str
            a list of column name for x variables
        y : str
            column name of the prediction
        layers : list of int
            The number of neurons in each dense hidden layer
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        res = self._preprocess(training, validation)
        self.mean_x, self.std_x, self.mean_y, self.std_y, self.x, \
            n_training, n_validation = res
        # create layers
        if restart:
            self._set_layers(layers, non_linear)
        # fitting
        self.train_loop(self._wrap(training), self._wrap(validation),
                        n_training, n_validation, **kwargs)

    def predict(self, X):
        """
        Make a prediction using the trained model.

        Parameters:
        -----------
        data : pandas.DataFrame

        Returns:
        --------
        Y : numpy.ndarray
        """
        X = X[self.x].values
        tensor = self._array_to_tensor(X, self.mean_x, self.std_x)
        with _torch.no_grad():
            output = self(tensor)
        Y = self._tensor_to_array(output, self.mean_y, self.std_y)
        return Y

    def forward(self, tensor):
        """Don't use this"""
        for i, dense in enumerate(self.dense_layers):
            tensor = dense(tensor)
        return tensor.view(-1)

    def _to_tensor(self, data):
        """Converts a dataframe into a tuple of tensors"""
        X, Y = data
        X, Y = X[self.x].values, _np.array(Y)
        X = self._array_to_tensor(X, mean=self.mean_x, std=self.std_x)
        Y = self._array_to_tensor(Y, mean=self.mean_y, std=self.std_y)
        return X, Y

    def _preprocess(self, training, validation):
        """preprocess the data"""
        # batchify if needed
        if not isinstance(training, _types.FunctionType):
            return self._preprocess(lambda: (yield training),
                                    lambda: (yield validation))
        # treat by batches
        sum_x, sum2_x, n_x = 0., 0., 0.
        sum_y, sum2_y, n_y = 0., 0., 0.
        x = None
        # treat training data
        for i, (X, Y) in enumerate(training()):
            if x is None:
                x = list(X.columns)
            for key in X.columns:
                if X[key].isna().any():
                    raise ValueError(f"At batch {i}: Found NaN in "
                                     f"training x['{key}']")
            if _np.isnan(Y).any():
                raise ValueError(f"At batch {i}: Found NaN in training y")
            sum_x, sum2_x, n_x = self._sums(X[x].values, sum_x, sum2_x, n_x)
            sum_y, sum2_y, n_y = self._sums(Y, sum_y, sum2_y, n_y)
        n_training = i+1
        # treat validation data
        for i, (X, Y) in enumerate(validation()):
            for key in X.columns:
                if X[key].isna().any():
                    raise ValueError(f"At batch {i}: Found NaN in "
                                     f"validation x['{key}']")
            if _np.isnan(Y).any():
                raise ValueError(f"At batch {i}: Found NaN in validation y")
            sum_x, sum2_x, n_x = self._sums(X.values, sum_x, sum2_x, n_x)
            sum_y, sum2_y, n_y = self._sums(Y, sum_y, sum2_y, n_y)
        n_validation = i+1
        # returns values
        mean_x, std_x = self._mean_std(sum_x, sum2_x, n_x)
        mean_y, std_y = self._mean_std(sum_y, sum2_y, n_y)
        return mean_x, std_x, mean_y, std_y, x, n_training, n_validation

    def _set_layers(self, *args):
        """Remove the previous layers and create new ones"""
        layers, non_linear = args
        self.args = args
        # Deleting previous layers
        for i, _ in enumerate(self.dense_layers):
            delattr(self, f"dense{i}")
        self.dense_layers = []
        # Set dense layers
        dens = _layers.DenseLayer
        n = len(self.x)
        for i, n_out in enumerate(layers):
            layer = dens(n, n_out, non_linear=non_linear)
            self.dense_layers.append(layer)
            setattr(self, f"dense{i}", layer)
            n = layer.shape_out
        # out layer
        i = len(self.dense_layers)
        n_out = 1
        layer = dens(n, n_out, non_linear="identity")
        self.dense_layers.append(layer)
        setattr(self, f"dense{i}", layer)
        # migrate to appropriate device
        self.to(self.device)
        # reset optimization
        self.reset()

    @property
    def dump(self):
        parameters = dict()
        parameters["x"] = self.x
        parameters["mean_x"] = self.mean_x.tolist()
        parameters["std_x"] = self.std_x.tolist()
        parameters["mean_y"] = self.mean_y.tolist()
        parameters["std_y"] = self.std_y.tolist()
        parameters["args"] = self.args
        dense = [d.dump for d in self.dense_layers]
        parameters["dense_layers"] = dense
        return {type(self).__name__: parameters, "version": _ml.__version__}

    @dump.setter
    def dump(self, other):
        if type(self).__name__ not in other.keys():
            raise ValueError(f"Expected dump from a '{type(self).__name__}'")
        other = other[type(self).__name__]
        self.x = other["x"]
        self.mean_x = _np.array(other["mean_x"])
        self.std_x = _np.array(other["std_x"])
        self.mean_y = _np.array(other["mean_y"])
        self.std_y = _np.array(other["std_y"])
        self.args = other["args"]
        if self.args is not None:
            self._set_layers(*self.args)
        for i, d in enumerate(other["dense_layers"]):
            self.dense_layers[i].dump = d
        self.reset_optimizer()

    @property
    def cpp(self):
        ny = len(self.std_y)
        code = "#include <vector>\n"
        code += "#include <math.h>\n"
        code += "double relu(double x) {if (x > 0.) {return x;} return 0.;}\n"
        code += "double tanh(double x) {return std::tanh(x);}\n"
        code += "double identity(double x) {return x;}\n\n"
        out_type = "std::vector<double>" if ny > 1 else "double"
        args = ", ".join(["double "+x for x in self.x])
        code += f"{out_type} neural_network({args})\n"
        code += "{\n"
        # Normalization
        for i, x in enumerate(self.x):
            code += f"\t{x} = ({x}-{self.mean_x[i]})/{self.std_x[i]};\n"
        # Layers
        inputs = self.x
        for i, layer in enumerate(self.dense_layers):
            for j, weights in enumerate(layer.linear.weight):
                code += f"\tdouble nn_{i}_{j} = {layer.non_linear}("
                parts = [None]*len(weights)
                for k, var in enumerate(inputs):
                    w = float(weights[k])
                    parts[k] = f"{inputs[k]}*{w}"
                if layer.linear.bias is not None:
                    parts.append(f"{layer.linear.bias[j]}")
                code += " + ".join(parts) + ");\n"
            inputs = [f"nn_{i}_{j}" for j, _ in enumerate(layer.linear.weight)]
        # normalization
        for k in range(ny):
            code += f"\tnn_{i}_{k} = nn_{i}_{k}*{self.std_y[k]}" \
                    f" + {self.mean_y[k]};\n"
        # return value
        if ny == 1:
            code += f"\treturn nn_{i}_{0};\n"
        else:
            code += "return {"
            code += ", ".join([f"nn_{i}_{k}" for k in range(ny)])
            code += "};\n"
        code += "}"
        return code



from ._decision_tree import RegressionTree, ClassificationTree
from machine_learning._templates import RegressorTemplate
from machine_learning._templates import ClassifierTemplate


class RandomForestRegressor(RegressorTemplate):
    pass


class RandomForestClassifier(ClassifierTemplate):
    pass



import torch
from typing import List
from .linear_stage import LinearStage


class FullyConnected(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump: dict) -> 'FullyConnected':
        obj = cls(1)
        obj.hidden_layers = torch.nn.ModuleList()
        obj.in_features = dump["in features"]
        for d in dump["hidden layers"]:
            obj.hidden_layers.append(LinearStage.from_dump(d))
        return obj

    def __init__(self, in_features: int,
                 hidden_layers: List[int] = [10, 10, 10],
                 activation: str = "relu"):
        super().__init__()
        self.in_features = in_features
        self.hidden_layers = torch.nn.ModuleList()
        for out_features in hidden_layers:
            lin = LinearStage(in_features, out_features, activation)
            self.hidden_layers.append(lin)
            in_features = out_features

    def forward(self, X: torch.Tensor) -> torch.tensor:
        for layer in self.hidden_layers:
            X = layer(X)
        return X

    def shape_in(self, shape_out: int):
        assert shape_out == self.out_channels
        return self.in_features

    def shape_out(self, shape_in: int):
        assert shape_in == self.in_channels
        return self.out_features

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "in features": self.in_features,
                "hidden layers": [h.dump for h in self.hidden_layers]}

    @property
    def out_features(self):
        if len(self.hidden_layers) > 0:
            return self.hidden_layers[-1].out_features
        else:
            return self.in_features



import ctypes as _c
import numpy as _np
from .._library_loader import _lib

# defining structures
c_double_p = _c.POINTER(_c.c_double)


class cBranch(_c.Structure):
    _fields_ = [("is_leaf", _c.c_bool),
                ("child_true", _c.c_uint),
                ("child_false", _c.c_uint),
                ("gain", _c.c_double),
                ("split_numerical_x", _c.c_bool),
                ("i", _c.c_uint),
                ("j", _c.c_uint),
                ("y_num", _c.c_double),
                ("iy_cat", _c.c_uint)]


class cTree(_c.Structure):
    _fields_ = [("n_branches", _c.c_uint),
                ("branches", _c.POINTER(cBranch))]


# delete tree
_lib.delete_tree.restype = None
_lib.delete_tree.argtype = [cTree]


def delete_tree(ctree: cTree):
    _lib.delete_tree(ctree)


# grow regressor_tree
_lib.grow_regressor.restype = cTree
_lib.grow_regressor.argtype = [c_double_p, _c.c_char_p, c_double_p, _c.c_uint,
                               _c.c_uint, _c.c_uint, _c.c_uint, _c.c_uint,
                               _c.c_uint, _c.c_uint, c_double_p]


def grow_regressor_tree(num_x: _np.ndarray, cat_x: _np.ndarray, y: _np.ndarray,
                        y_pred: _np.ndarray = None,
                        min_samples: int = 1, max_depth: int = 1000,
                        max_leafs: int = 1000):
    assert isinstance(num_x, _np.ndarray)
    assert isinstance(cat_x, _np.ndarray)
    assert isinstance(y, _np.ndarray)
    assert cat_x.dtype.char == "U" or cat_x.dtype.char == "S"
    n_rows = len(y)
    _, n_numerical_columns = num_x.shape
    _, n_categorical_columns = cat_x.shape
    n_chars_x = cat_x.dtype.itemsize
    num_x = num_x.ctypes.data_as(c_double_p)
    cat_x = cat_x.ctypes.data_as(_c.c_char_p)
    y = y.ctypes.data_as(c_double_p)
    if isinstance(y_pred, _np.ndarray):
        y_pred = y_pred.ctypes.data_as(c_double_p)
    else:
        y_pred = None
    return _lib.grow_regressor(num_x, cat_x, y, n_rows, n_numerical_columns,
                               n_categorical_columns, n_chars_x, min_samples,
                               max_depth, max_leafs, y_pred)


# grow classifier_tree
_lib.grow_regressor.restype = cTree
_lib.grow_regressor.argtype = [c_double_p, _c.c_char_p, _c.c_char_p, _c.c_uint,
                               _c.c_uint, _c.c_uint, _c.c_uint, _c.c_uint,
                               _c.c_uint, _c.c_uint, _c.c_uint, _c.c_char_p]


def grow_classifier_tree(num_x: _np.ndarray, cat_x: _np.ndarray,
                         y: _np.ndarray,
                         min_samples: int = 1, max_depth: int = 1000,
                         max_leafs: int = 1000):
    assert isinstance(num_x, _np.ndarray)
    assert isinstance(cat_x, _np.ndarray)
    assert isinstance(y, _np.ndarray)
    assert cat_x.dtype.char == "U" or cat_x.dtype.char == "S"
    assert y.dtype.char == "U" or y.dtype.char == "S"
    n_rows = len(y)
    _, n_numerical_columns = num_x.shape
    _, n_categorical_columns = cat_x.shape
    n_chars_x = cat_x.dtype.itemsize
    n_chars_y = y.dtype.itemsize
    num_x = num_x.ctypes.data_as(c_double_p)
    cat_x = cat_x.ctypes.data_as(_c.c_char_p)
    y = y.ctypes.data_as(_c.c_char_p)
    return _lib.grow_classifier(num_x, cat_x, y, n_rows, n_numerical_columns,
                                n_categorical_columns, n_chars_x, n_chars_y,
                                min_samples, max_depth, max_leafs, None)



import types as _types
import math as _math
import torch as _torch
import numpy as _np
import matplotlib.pyplot as _plt
import machine_learning as _ml
import machine_learning.neural_networks.layers as _layers
import machine_learning._utilities as _utils


functions = {'relu': _torch.nn.functional.relu,
             'tanh': _torch.tanh,
             'sigmoid': _torch.sigmoid,
             'hard tanh': _torch.nn.functional.hardtanh,
             'log sigmoid': _torch.nn.functional.logsigmoid,
             'identity': lambda x: x}


def as_function(non_linear):
    """
    Returns a function corresponding to the given name.
    If non_linear is already a function, returns it.
    """
    if isinstance(non_linear, _types.FunctionType):
        return non_linear
    if non_linear not in functions.keys():
        raise KeyError(f"got {non_linear} for 'non_linear'"
                       f", expected one of {tuple(functions.keys())}")
    return functions[non_linear]


def as_function_name(non_linear):
    """returns the name corresponding to the function"""
    if not isinstance(non_linear, _types.FunctionType):
        raise TypeError(f"Expected a function but got '{type(non_linear)}'")
    for name, func in functions.items():
        if func == non_linear:
            return name
    else:
        return "unknown"


class NNtemplate(_torch.nn.Module):
    """
    A template of neural network model

    Attributes:
    -----------
    GPU : bool
        Whether to use GPU or not
    history : dict
        dictionary with history of fitting
    loss : function
        the loss(output, target) function
    optimizer : torch.optim.Optimizer
        the optimizer
    L2 : float
        L2 normalization factor
    learning_rate : float
        The learning rate of the optimization
    fitted : bool
        Whether the model was fitted already
    """

    def __init__(self):
        """initialize default attributes"""
        super().__init__()
        self.GPU = _torch.cuda.is_available()
        self.history = self._empty_history()
        self.loss = _torch.nn.functional.mse_loss
        self._L2 = 0.
        self._L1 = 0.
        self._learning_rate = 1.0E-3
        dummy = [_torch.nn.Parameter(_torch.tensor([1.]))]
        self.optimizer = _torch.optim.Adam(dummy, weight_decay=self.L2,
                                           lr=self.learning_rate)
        self.fitted = False

    def train_loop(self, training, validation, n_training, n_validation,
                   n_epochs=100, learning_rate=1.0E-3, L1=0., L2=0.,
                   keep_best=True, patience=None, verbose=True):
        """
        train using batch gradient descent on provided mini-batchs

        Parameters:
        -----------
        training : generator function
            a function that yield (input, target)
        validation : generator function
            a function that yields (input, target)
        n_training : int
            number of batchs yielded by 'training'
        n_validation : int
            number of batchs yielded by 'validation'

        Other parameters:
        -----------------
        n_epochs : int
            number of epoch to perform on training data
        learning_rate : float
            Learning rate of the model
        keep_best : bool
            If True, keep the epoch of best validation loss
        patience : int
            Stops training if validation loss has not decreased in
            'patience' epochs
        GPU : bool
            If True, trains on GPU
        verbose : bool
            If True, prints during fitting
        """
        # activate training mode
        self.train()
        # If a tuple of itterables is passed, transform them into a generator
        if not isinstance(training, _types.FunctionType):
            return self.train_loop(lambda: (yield training),
                                   lambda: (yield validation),
                                   1, 1, n_epochs=n_epochs,
                                   learning_rate=learning_rate, L1=L1, L2=L2,
                                   keep_best=keep_best, patience=patience,
                                   verbose=verbose)
        # setting hyper-parameters
        _torch.backends.cudnn.benchmark = True
        self.learning_rate = learning_rate
        self.L2 = L2
        self.L1 = L1
        # loading data
        best_params = self.state_dict()
        optim_state = self.optimizer.state_dict()
        training_loss, validation_loss = [], []
        best_epoch = self.history["best epoch"]
        if best_epoch is not None:
            best_loss = self.history["validation loss"][best_epoch]
        else:
            best_loss = float("inf")
        # Stoping training if user press ctrl+c
        try:
            # Looping on epochs
            k = 0 if best_epoch is None else best_epoch+1
            for epoch in range(k, k+n_epochs):
                if verbose:
                    print(f"Epoch {epoch} : ", end="")
                # looping on training mini-batchs
                mean_loss = []
                self.optimizer.zero_grad()
                for batch, (input, target) in enumerate(training()):
                    output = self(input)
                    loss = self.loss(output, target)
                    (loss/n_training + self._L1_reg()).backward()
                    mean_loss.append(loss.item())
                    # printing progress
                    if verbose:
                        self._print_progress(batch, n_training)
                self.optimizer.step()
                training_loss.append(_np.mean(mean_loss))
                if batch+1 != n_training:
                    raise ValueError(f"training yielded {batch+1} batchs, "
                                     f"but {n_training} were expected")
                # looping on validation mini-batchs
                mean_loss = []
                for batch, (input, target) in enumerate(validation()):
                    with _torch.no_grad():
                        output = self(input)
                        loss = self.loss(output, target)
                        mean_loss.append(loss.item())
                validation_loss.append(_np.mean(mean_loss))
                # printing progress
                if verbose:
                    print(f" training={training_loss[-1]:.3g}"
                          f" validation={validation_loss[-1]:.3g}")
                # Saving state
                if (validation_loss[-1] < best_loss) or not keep_best:
                    best_epoch = epoch
                    best_loss = validation_loss[-1]
                    best_params = self.state_dict()
                    optim_state = self.optimizer.state_dict()
                # Interupt training if no improvement
                if not(patience is None) and (epoch - best_epoch > patience):
                    break
        except KeyboardInterrupt:
            pass
        if verbose:
            print()
        # Returning results
        self.fitted = True
        self.history["training loss"].extend(training_loss)
        self.history["validation loss"].extend(validation_loss)
        self.history["best epoch"] = best_epoch if keep_best else epoch
        # Loading saved state
        if keep_best:
            self.load_state_dict(best_params)
            self.optimizer.load_state_dict(optim_state)
        # activate evaluation mode
        self.eval()

    def plot_history(self, ax=None, log=False):
        """
        Plot the history of the fitting
        """
        if ax is None:
            _plt.figure()
            ax = _plt.subplot(1, 1, 1)
        N = min(len(self.history["training loss"]),
                len(self.history["validation loss"]))
        epochs = [i for i in range(0, N)]
        if not self.history["best epoch"] is None:
            ax.axvline(x=self.history["best epoch"], color="k")
        ax.plot(epochs, self.history["training loss"][0:N], linestyle="",
                marker=".", color="C0",
                label="training loss (batchs average)")
        ax.plot(epochs, self.history["validation loss"][0:N], linestyle="",
                marker=".", color="C1",
                label="validation loss (batchs average)")
        if log:
            ax.set_yscale("log")
        ax.set_xlabel("Epochs")
        ax.set_ylabel("Loss")
        ax.legend()
        ax.get_figure().tight_layout()
        return ax

    def seed(self, seed):
        """
        Initialize the seed for the random numbers generation.
        This calls numpy.random.seed and torch.manuel_seed

        Kwargs:
            seed : int or None
                The seed to use
        """
        if seed is not None:
            _np.random.seed(seed)
            _torch.manual_seed(seed)
            _torch.backends.cudnn.deterministic = True
            _torch.backends.cudnn.benchmark = False

    def reset(self):
        """
        Reset the model to an untrained state.

        Calls reset optimizer under the hood.
        """
        self.history = self._empty_history()
        self.fitted = False
        # reset parameters
        for module in self.children():
            if hasattr(module, "reset_parameters"):
                module.reset_parameters()
        # reset optimizer
        self.reset_optimizer()

    def reset_optimizer(self):
        """
        Reset the optimizer.

        If some parameters are added to or deleted from the model,
        this must be called. Otherwise optimizer won't aknowledge it.
        """
        opt_class = type(self.optimizer)
        if next(self.parameters(), None) is None:
            parameters = [_torch.nn.Parameter(_torch.tensor([1.]))]
        else:
            parameters = self.parameters()
        self.optimizer = opt_class(parameters,
                                   weight_decay=self.L2,
                                   lr=self.learning_rate)

    def forward(self, input):
        """Implement me"""
        raise Exception("Not implemented")

    def _print_progress(self, batch, n_training):
        """print progress during 'train_loop_batch' """
        now = batch/n_training
        after = (batch+1)/n_training
        k_now = _math.floor(10*now)
        k_after = _math.floor(10*after)
        for i in range(k_now, k_after):
            print(u"\u2588", end="")

    def _empty_history(self):
        """default value for the fitting history"""
        return {"training loss": [],
                "validation loss": [],
                "best epoch": None}

    def _dataframe_to_arrays(self, df, x, y, validation=None):
        """converts dataframe data to training and validation"""
        training = df[x].values, df[y].values
        if df is not None:
            validation = validation[x].values, validation[y].values
        return training, validation

    def _validation_split(self, training, validation, validation_fraction=0.2):
        """
        if validation is None, split training data
        into training and validation

        Parameters:
        -----------
        training : tuple of iterables
            tuple of X, Y for training
        validaton : tuple of iterables, or None
            tuple of X, Y for validation
        validation_fraction : float
            fraction of validation data if validation is None

        Returns:
        --------
            training, validation : tuple of iterables
        """
        if validation is None:
            X, Y = training
            assert len(X) == len(Y), "X and Y must have the same dimensions to perform validation split"
            n_total = len(X)
            index = _np.random.permutation(n_total)
            n_validation = int(validation_fraction*n_total)
            if n_validation == 0:
                raise ValueError("Validation split led to 0 validation data")
            if n_validation == n_total:
                raise ValueError("Validation split led to 0 training data")
            X_val = _utils.subset(X, index[:n_validation])
            Y_val = _utils.subset(Y, index[:n_validation])
            X_train = _utils.subset(X, index[n_validation:])
            Y_train = _utils.subset(Y, index[n_validation:])
            training = X_train, Y_train
            validation = X_val, Y_val
        return training, validation

    def _sums(self, X, sum_=0., sum2=0., n=0.):
        """calculate sum and sum of squares of X. Used for normalization"""
        if (isinstance(X, list) or
           isinstance(X, tuple)) and isinstance(X[0], _ml.Image):
            X = _np.array([image.tensor for image in X])
        else:
            X = _np.array(X)
        n += _np.product([k for i, k in enumerate(X.shape) if i != 1])
        axis = tuple([i for i in range(0, len(X.shape)) if i != 1])
        shape = tuple([-1]+[1 for _ in X.shape[2:]])
        sum_ += _np.sum(X, axis=axis).reshape(*shape)
        sum2 += _np.sum(X**2, axis=axis).reshape(*shape)
        return sum_, sum2, n

    def _mean_std(self, sum_, sum2, n):
        """calculate mean and std from sums"""
        mean = sum_ / n
        std = _np.sqrt(sum2 / n - mean**2)
        return mean, std

    def _images_to_tensor(self, images, mean=0., std=1.):
        """converts an image to a normalized tensor"""
        if isinstance(images, _ml.Image):
            images = [images]
        data = [(image.tensor-mean)/std for image in images]
        return _torch.tensor(data, dtype=_torch.float, device=self.device)

    def _tensor_to_images(self, tensor, mean=0., std=1.):
        """converts a normalized tensor to an image"""
        tensor = tensor.cpu().detach().numpy()
        tensor = tensor*std + mean
        images = [_ml.Image(tensor=data) for data in tensor]
        return images

    def _array_to_tensor(self, arrays, mean=0., std=1.):
        """converts an array to a normalized tensor"""
        data = (arrays-mean)/std
        return _torch.tensor(data, dtype=_torch.float, device=self.device)

    def _tensor_to_array(self, tensor, mean=0., std=1.):
        """converts a normalized tensor to array"""
        array = tensor.cpu().detach().numpy()
        array = array*std + mean
        return array

    def _labels_to_tensor(self, labels):
        Y = [self.categories.index(label) for label in labels]
        return _torch.tensor(Y, dtype=_torch.long, device=self.device)

    def _tensor_to_labels(self, tensor):
        indexes = _torch.argmax(tensor, dim=1)
        labels = [self.categories[index] for index in indexes]
        return labels

    def _segmented_images_to_tensor(self, segmented):
        """convert grayscale segmented images to a tensor"""
        S = [image.data for image in segmented]
        Y = [_np.argmax([_np.all(s == c, axis=-1)
             for c in self.categories], axis=0)
             for s in S]
        return _torch.tensor(Y, dtype=_torch.long, device=self.device)

    def _tensor_to_segmented_images(self, tensor):
        data = tensor.detach().cpu().numpy()
        data = _np.argmax(data, axis=1)
        data = _np.array(self.categories)[data]
        return [_ml.Image(data=d) for d in data]

    def _to_tensor(self, data):
        """converts a tuple (X, Y) into a tuple of tensors"""
        raise NotImplementedError("'_to_tensor' is called but not implemented")

    def _wrap(self, generator):
        """return a generator of tuple of tensors from a generator of (X, Y)"""
        def wrapper():
            for data in generator():
                yield self._to_tensor(data)
        return wrapper

    def _wrap_batch(self, data, n_max=None):
        """
        wrap a tuple of tensors into a generator yielding batchs
        of at most n_max observations

        parameters:
        -----------
        data : tuple of tensors
            the training or validation data
        n_max : int or None
            the maximum number of observation in a single batch
        """
        X, Y = data
        if n_max is not None:
            n_batchs = _math.ceil(len(X)/n_max)
        else:
            n_batchs = 1

        def wrapper():
            step = len(X)/n_batchs
            stop = 0
            for i in range(1, n_batchs+1):
                start = stop
                stop = int(i*step)
                yield X[start:stop], Y[start:stop]
        return wrapper, n_batchs

    def _L1_reg(self):
        """Returns the L1 regularization term of the neural network
        (L1 norm of all parameters times the factor)"""
        if self.L1 == 0.:
            return 0.
        l1_n = [(p.abs().sum(), _np.product(p.shape))
                for p in self.parameters()]
        l1, n = zip(*l1_n)
        mean_l1_norm = sum(l1)/sum(n)
        return self.L1*mean_l1_norm

    @property
    def GPU(self):
        return self._GPU

    @GPU.setter
    def GPU(self, other):
        self._GPU = other
        assert not(self.GPU) or _torch.cuda.is_available(), \
            "CUDA is not available on this machine. Unable to run on GPU."
        self.to(self.device)

    @property
    def device(self):
        """returns the device the model should train on"""
        if self.GPU:
            return _torch.device("cuda:0")
        else:
            return _torch.device("cpu")

    @property
    def learning_rate(self):
        return self._learning_rate

    @learning_rate.setter
    def learning_rate(self, other):
        self._learning_rate = other
        for group in self.optimizer.param_groups:
            group["lr"] = other

    @property
    def L2(self):
        return self._L2

    @L2.setter
    def L2(self, other):
        self._L2 = other
        for group in self.optimizer.param_groups:
            group["weight_decay"] = other

    @property
    def VRAM(self):
        """current VRAM usage (allocated and cached in capability %)"""
        t = _torch.cuda.get_device_properties(self.device).total_memory
        a = _torch.cuda.memory_allocated(self.device)
        allocated = a/t * 100.
        return allocated

    @property
    def VRAM_peak(self):
        """peak allocation % of VRAM since last call"""
        t = _torch.cuda.get_device_properties(self.device).total_memory
        a = _torch.cuda.max_memory_allocated(self.device)
        _torch.cuda.reset_max_memory_allocated(self.device)
        allocated = a/t * 100.
        return allocated


class LayerActivationWatcher():
    """
    visualize the intermediate results of a neural network.
    For deep dreaming applications.

    Parameters:
    -----------
    activation : torch.tensor or None
        activation of the target layer at last call.
    """
    def __init__(self, layer):
        self.activation = None
        if isinstance(layer, _layers.ConvolutionLayer):
            module = layer.convolution
        elif isinstance(layer, _layers.DenseLayer):
            module = layer.linear
        else:
            raise NotImplementedError("DeepDream is not implemented for "
                                      f"layer '{type(layer)}'")
        self.hook = module.register_forward_hook(self.hook_function)

    def __del__(self):
        self.hook.remove()

    def hook_function(self, module, input, output):
        self.activation = output



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
# from .utilities import MSE, RMSE, R2, accuracy, plot_correlation


def regressor(cls):

    class Regressor(cls):

        def __init__(self, *args, **kwargs):
            """
            """
            super().__init__(*args, **kwargs)

        def plot_correlation(self, x, y, weights=None, ax=None, label="_"):
            """
            """
            y_pred = self(x)
            ax.scatter(y, y_pred, marker=".", s=weights, label=label)
            points = np.concatenate([c.get_offsets() for c in ax.collections])
            inf, sup = points.min(), points.max()
            delta = sup - inf if sup != inf else 1
            sup += 0.01*delta
            inf -= 0.01*delta
            plt.plot([inf, sup], [inf, sup], color="k", zorder=0)
            ax.set_xlim([inf, sup])
            ax.set_ylim([inf, sup])
            ax.set_xlabel("target")
            ax.set_ylabel("predicted")
            ax.set_aspect("equal", "box")
            ax.legend()

    return Regressor


def classifier(cls):

    class Classifier(cls):

        def __init__(self, *args, **kwargs):
            """
            """
            super().__init__(*args, **kwargs)

        def plot_confusion_matrix(self, x, y, ax=None, cmap="Blues"):
            """
            """
            indexes = range(len(self.categories))
            if ax is None:
                f, ax = plt.subplots()
            y_pred = pd.Series(self.index(x))
            y_target = pd.Series([self.categories.index(c) for c in y])
            tab = pd.crosstab(y_pred, y_target, normalize="all")
            for i in indexes:
                if i not in tab.index:
                    tab.loc[i] = 0
                if i not in tab.columns:
                    tab[i] = 0
            tab = tab.to_numpy()
            ax.imshow(tab, origin="lower", interpolation="nearest", cmap=cmap)
            ax.grid(False)
            ax.set_xticks(indexes)
            ax.set_xticklabels(self.categories)
            ax.set_xlabel("target")
            ax.set_yticks(indexes)
            ax.set_yticklabels(self.categories)
            ax.set_ylabel("predicted")
            for y in indexes:
                for x in indexes:
                    val = tab[y, x]
                    if val >= 0.01:
                        ax.text(x, y, f"{val:.2f}", va='center', ha='center')

    return Classifier



import torch
import torch.nn.functional as F
from typing import Union, List, Tuple
from .pooling_stage import PoolingStage2d, PoolingStage1d, load_poolingstage


class _Encoder(torch.nn.Module):
    """
    An encoder is a succession of PoolingStage
    It reduces spatial dimensions of an image but increase channels depth
    """

    @classmethod
    def from_dump(cls, dump: dict) -> '_Encoder':
        obj = cls(1)
        obj.stages = torch.nn.ModuleList()
        for d in dump["stages"]:
            obj.stages.append(load_poolingstage(d))
        return obj

    def __init__(self):
        super().__init__()

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        for stage in self.stages:
            X = stage(X)
        return X

    def shape_out(self, shape_in: List[int]) -> List[int]:
        shape = shape_in
        for stage in self.stages:
            shape = stage.shape_out(shape)
        return shape

    def shape_in(self, shape_out: List[int]) -> List[int]:
        shape = shape_out
        for stage in self.stages[::-1]:
            shape = stage.shape_in(shape)
        return shape

    def shapes(self, shape_in: List[int]) -> List[List[int]]:
        shape = shape_in
        shapes_list = []
        for stage in self.stages:
            shape = stage.shape_out(shape)
            shapes_list.append(shape)
        return shapes_list

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "stages": [s.dump for s in self.stages]}

    @property
    def in_channels(self):
        return self.stages[0].in_channels

    @property
    def out_channels(self):
        return self.stages[-1].out_channels


class Encoder1d(_Encoder):

    def __init__(self, in_channels: int,
                 convolution_stages: List[Union[dict, List[dict]]] = [],
                 pooling_windows: List[int] = [],
                 pooling_type: str = "Max",
                 padded: bool = True,
                 stacked: bool = False,
                 activation: str = "relu"):
        """
        in_channels : int
            The number of channels of the input
        convolution_stages : list of dict, or list of list of dict
            the kwargs used to build the successive Convolution1dStage before
            each pooling.
        pooling_windows : list of int
            the window size of the pooling layers
            can be omited if pooling_type is None
        pooling_type : one of {None, "Max", "Avg"}
            The type of pooling to perform
        padded: bool
            default value for the "padded" key in the kwargs
            If omited, the default is used
        stacked: bool
            default value for the "stacked" key in the kwargs
            If omited, the default is used
        activation: str
            default value for the "activation" key in the kwargs
            If omited, the default is used
        """
        super().__init__()
        self.stages = torch.nn.ModuleList()
        for conv, pool in zip(convolution_stages, pooling_windows):
            conv.setdefault("padded", padded)
            conv.setdefault("activation", activation)
            cp = PoolingStage1d(in_channels, conv,
                                pooling_type=pooling_type,
                                pooling_size=pool,
                                padded=padded,
                                stacked=stacked,
                                activation=activation)
            self.stages.append(cp)
            in_channels = cp.out_channels


class Encoder2d(_Encoder):

    def __init__(self, in_channels: int,
                 convolution_stages: List[Union[dict, List[dict]]] = [],
                 pooling_windows: List[Tuple[int, int]] = [],
                 pooling_type: str = "Max",
                 padded: bool = True,
                 stacked: bool = False,
                 activation: str = "relu"):
        """
        in_channels : int
            The number of channels of the input
        convolution_stages : list of dict, or list of list of dict
            the kwargs used to build the successive Convolution1dStage before
            each pooling.
        pooling_windows : list of tuple of int
            the (height, width) window size of the pooling layers
            can be omited if pooling_type is None
        pooling_type : one of {None, "Max", "Avg"}
            The type of pooling to perform
        padded: bool
            default value for the "padded" key in the kwargs
            If omited, the default is used
        stacked: bool
            default value for the "stacked" key in the kwargs
            If omited, the default is used
        activation: str
            default value for the "activation" key in the kwargs
            If omited, the default is used
        """
        super().__init__()
        self.stages = torch.nn.ModuleList()
        for conv, pool in zip(convolution_stages, pooling_windows):
            cp = PoolingStage2d(in_channels, conv,
                                pooling_type=pooling_type,
                                pooling_size=pool,
                                padded=padded,
                                stacked=stacked,
                                activation=activation)
            self.stages.append(cp)
            in_channels = cp.out_channels



from .._library_loader import _lib
import ctypes as _c
import numpy as _np

# Typing of the Tensor functions
_lib.new_Tensor.restype = _c.c_void_p
_lib.new_Tensor.argtypes = [_c.POINTER(_c.c_double),
                            _c.POINTER(_c.c_uint),
                            _c.c_uint]

_lib.Tensor_copy.restype = _c.c_void_p
_lib.Tensor_copy.argtypes = [_c.c_void_p]

_lib.Tensor_reshape.restype = _c.c_void_p
_lib.Tensor_reshape.argtypes = [_c.c_void_p,
                                _c.POINTER(_c.c_uint),
                                _c.c_uint]

_lib.Tensor_ndims.restype = _c.c_uint
_lib.Tensor_ndims.argtypes = [_c.c_void_p]

_lib.Tensor_shape.restype = _c.POINTER(_c.c_uint)
_lib.Tensor_shape.argtypes = [_c.c_void_p]

_lib.Tensor_data.restype = _c.POINTER(_c.c_double)
_lib.Tensor_data.argtypes = [_c.c_void_p]

_lib.Tensor_set.restype = None
_lib.Tensor_set.argtypes = [_c.c_void_p, _c.POINTER(_c.c_uint), _c.c_double]

_lib.Tensor_get.restype = _c.c_double
_lib.Tensor_get.argtypes = [_c.c_void_p, _c.POINTER(_c.c_uint)]

_lib.Tensor_subTensor.restype = _c.c_void_p
_lib.Tensor_subTensor.argtypes = [_c.c_void_p, _c.c_uint]

_lib.del_Tensor.restype = None
_lib.del_Tensor.argtypes = [_c.c_void_p]

_lib.Tensor_concatenate.restype = _c.c_void_p
_lib.Tensor_concatenate.argtypes = [_c.POINTER(_c.c_void_p),
                                    _c.c_uint, _c.c_uint]


# Defining the class
class Tensor:

    def concatenate(tensors: list, axis: int = 0) -> 'Tensor':
        assert axis >= 0
        assert len(tensors) > 0
        arr = (_c.c_void_p * len(tensors))(*[t.pointer for t in tensors])
        return Tensor(_c.c_void_p(_lib.Tensor_concatenate(arr,
                                                          len(tensors), axis)))

    def __init__(self, data):
        if isinstance(data, _c.c_void_p):
            self.pointer = data
        elif hasattr(data, "__iter__") and not isinstance(data, str):
            arr = _np.array(data, dtype=_np.double, order="C")
            shape = arr.shape
            ndims = len(shape)
            shape = (_c.c_uint * ndims)(*shape)
            p = arr.ctypes.data_as(_c.POINTER(_c.c_double))
            self.pointer = _lib.new_Tensor(p, shape, ndims)
        else:
            # Set the poinetr to nullptr so that the finalizer '__del__'
            # doesn't raise an error
            self.pointer = None
            # Then raise an error
            raise TypeError(f"Unexpected type '{type(data)}'"
                            "in Tensor's constructor")

    def __del__(self):
        _lib.del_Tensor(self.pointer)

    def __getitem__(self, key):
        S = self.shape
        if (type(key) == int) and (len(S) == 1):
            key = (key,)
        T = type(key)
        if T == int:
            if (key >= S[0]) or (key < 0):
                raise IndexError("Indexing tensor out of range")
            p = _c.c_void_p(_lib.Tensor_subTensor(self.pointer, key))
            return Tensor(p)
        elif T == tuple:
            if len(key) != len(S):
                raise IndexError("Index size and tensor shape dont match")
            index = (_c.c_uint*len(S))(*key)
            return _lib.Tensor_get(self.pointer, index)
        else:
            raise TypeError(f"Unexpected index type '{T}'")

    def __setitem__(self, key, value):
        S = self.shape
        if not hasattr(key, "__iter__"):
            key = (key,)
        if len(key) != len(S):
            raise IndexError("Index size and tensor shape don't match")
        assert [isinstance(k, int) for k in key]
        assert [k >= 0 and k < s for k, s in zip(key, S)]
        index = (_c.c_uint * len(key))(*key)
        _lib.Tensor_set(self.pointer, index, value)

    def __repr__(self):
        return f"Tensor(address={self.pointer}, shape={self.shape})"

    def copy(self):
        """Returns a deep copy of the tensor"""
        return Tensor(_c.c_void_p(_lib.Tensor_copy(self.pointer)))

    def reshape(self, new_shape: list):
        """Returns the same tensor reshaped (underlying data is shared)"""
        assert _np.product(new_shape) == _np.product(self.shape)
        ndims = len(new_shape)
        new_shape = (_c.c_uint * ndims)(*new_shape)
        return Tensor(_c.c_void_p(_lib.Tensor_reshape(self.pointer,
                                                      new_shape, ndims)))

    def flatten(self):
        """Returns the tensor reshaped as a 1D tensor"""
        return self.reshape([_np.product(self.shape)])

    @property
    def numpy(self):
        """Returns the tensor as a numpy array (memory is not shared)"""
        S = self.shape
        address = _lib.Tensor_data(self.pointer)
        c_array = _c.cast(address, _c.POINTER(_c.c_double*_np.product(S)))[0]
        return _np.array(c_array, order="C").reshape(S)

    @property
    def shape(self):
        """size of each dimension of the tensor"""
        _shape = _lib.Tensor_shape(self.pointer)
        return [_shape[i] for i in range(self.ndims)]

    @property
    def ndims(self):
        """number of dimensions of the tensor"""
        return _lib.Tensor_ndims(self.pointer)



from .activation import Activation
from .batch_norm import BatchNorm1d, BatchNorm2d
from .convolution import Conv1d, Conv2d
from .linear import Linear
from .padding import ConstantPad1d, ConstantPad2d
from .pooling import MaxPool1d, MaxPool2d, AvgPool1d, AvgPool2d
from .pooling import OverallPool1d, OverallPool2d
from .linear_stage import LinearStage
from .convolution_stage import ConvStage1d, ConvStage2d
from .pooling_stage import PoolingStage1d, PoolingStage2d
from .encoder import Encoder1d, Encoder2d
from .u_net import UNet1d, UNet2d
from .fully_connected import FullyConnected



import torch
import torch.nn.functional as F


def linear_interpolation(X: torch.Tensor, size: tuple) -> torch.Tensor:
    """
    Returns a linear interpolation of the input tensor

    Parameters
    ----------
    X : torch.Tensor
        A tensor of shape (N, C, L)
    size : tuple of int
        The output size (L_out)

    Returns
    -------
    torch.Tensor :
        The interpolated tensor of shape (N, C, L_out)
    """
    assert len(X.shape) == 3
    return F.interpolate(X, size=size, mode="linear")


def bilinear_interpolation(X: torch.Tensor, size: tuple) -> torch.Tensor:
    """
    Returns a bilinear interpolation of the input tensor

    Parameters
    ----------
    X : torch.Tensor
        A tensor of shape (N, C, H, W)
    size : tuple of int
        The output size (H_out, W_out)

    Returns
    -------
    torch.Tensor :
        The interpolated tensor of shape (N, C, H_out, W_out)
    """
    assert len(X.shape) == 4
    return F.interpolate(X, size=size, mode="bilinear", align_corners=False)


def overall_pool(X: torch.Tensor, pooling_type: str = "Max"):
    """
    Apply over-all pooling to keep
    """
    dims = X.shape[2:]
    if any(d != 1 for d in dims):
        func = getattr(F, pooling_type.lower()+"_pool2d")
        X = func(X, dims)
    return X.view(X.shape[0], -1)



import torch as _torch
import machine_learning.neural_networks.layers as _layers
from machine_learning.neural_networks._regressor import Regressor


# Old monotonic method
class Monotonic(Regressor):
    """Penalization for non Monotonic variations"""
    def __init__(self, **kwargs):
        super(Monotonic, self).__init__(**kwargs)
        self.alpha = 1.
        self.sign = self._as_tensor(1.)
        
    def fit(self, *args, alpha=None, sign=None, **kwargs):
        if not alpha is None:
            self.alpha = alpha
        if not sign is None:
            self.sign = self._as_tensor(sign)
        return Regressor.fit(self, *args, **kwargs)
        
    def fit_batch(self, *args, alpha=None, sign=None, **kwargs):
        if not alpha is None:
            self.alpha = alpha
        if not sign is None:
            self.sign = self._as_tensor(sign)
        return Regressor.fit_batch(self, *args, **kwargs)
    
    def _grad(self, tensor):
        """
        Computes the grad of the prediction with respect to the input data
        This function zeros the gradient of the model's parameters as a side effect
        """
        tensor = tensor.clone().detach().requires_grad_()
        with _torch.enable_grad():
            prediction = self(tensor)
            prediction.sum().backward(create_graph=True)
        grad = tensor.grad
        self.optimizer.zero_grad()
        return grad
    
    def _penalization(self, grad):
        """
        Computes a penalization term for negative gradients, so that the 
        model be monotonically increasing
        """
        grad = grad*self.sign
        negative_grads = _torch.clamp_max(grad, 0.)#only keep the negative gradients
        #This is fidel to the publication
        #normed_grads = _torch.mean(_torch.min(negative_grads, dim=1)[0]**2)
        normed_grads = _torch.mean(_torch.norm(negative_grads, dim=1))
        return self.alpha * normed_grads
    
    def _get_loss_tensor_(self, X, Y):
        """
        compute the loss tensor
        Override the template so that a penalization for negative 
        gradients of the function is added to the loss
        """
        tensor,truth = self._transform_XY(X,Y)
        grad = self._grad(tensor)
        prediction = self(tensor)
        loss = self.loss(prediction,truth) + self._penalization(grad)
        return loss

class Gated(Regressor):
    """The weights of the linear gates are always positive: function always raising"""
    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        self.y_min = None

    @property
    def hidden_layers(self):
        return self._hidden_layers
    
    @hidden_layers.setter
    def hidden_layers(self, other):
        """Generate the hidden layers"""
        if (self._hidden_layers == tuple(other)) and self.fitted:#we dont reset the shape if it is not needed
            return
        self._hidden_layers = tuple(other)
        for i,layer in enumerate(self.layers):
            delattr(self,f"layer{i}")
        self.layers = []
        new_shape = self.shape
        for i,_ in enumerate(new_shape):
            if i == len(new_shape)-1:
                continue
            layer = _layers.LinearPW(new_shape[i],new_shape[i+1])
            #layer = _torch.nn.Linear(new_shape[i],new_shape[i+1])
            setattr(self, f"layer{i}", layer)
            self.layers.append(layer)
        self.reset_optimization()


from ..._library_loader import _lib
from ...tensor import Tensor as _Tensor
import ctypes as _c
import json as _json

# Typing of the Tensor functions
_lib.new_ConvolutionLayer.restype = _c.c_void_p
_lib.new_ConvolutionLayer.argtypes = [_c.c_char_p]

_lib.apply_ConvolutionLayer.restype = _c.c_void_p
_lib.apply_ConvolutionLayer.argtypes = [_c.c_void_p, _c.c_void_p]

_lib.del_ConvolutionLayer.restype = None
_lib.del_ConvolutionLayer.argtypes = [_c.c_void_p]

_lib.new_UpsamplingLayer.restype = _c.c_void_p
_lib.new_UpsamplingLayer.argtypes = [_c.c_char_p]

_lib.apply_UpsamplingLayer.restype = _c.c_void_p
_lib.apply_UpsamplingLayer.argtypes = [_c.c_void_p, _c.c_void_p, _c.c_void_p]

_lib.del_UpsamplingLayer.restype = None
_lib.del_UpsamplingLayer.argtypes = [_c.c_void_p]

_lib.new_DenseLayer.restype = _c.c_void_p
_lib.new_DenseLayer.argtypes = [_c.c_char_p]

_lib.apply_DenseLayer.restype = _c.c_void_p
_lib.apply_DenseLayer.argtypes = [_c.c_void_p, _c.c_void_p]

_lib.del_DenseLayer.restype = None
_lib.del_DenseLayer.argtypes = [_c.c_void_p]


# Defining the classes
class ConvolutionLayer:

    def __init__(self, data):
        if isinstance(data, _c.c_void_p):
            self.pointer = data
        elif isinstance(data, dict):
            settings = _json.dumps(data)
            self.pointer = _lib.new_ConvolutionLayer(settings.encode("utf-8"))
        else:
            raise TypeError(f"Unexpected type '{type(data)}'"
                            "in ConvolutionLayer's constructor")

    def __call__(self, input):
        assert isinstance(input, _Tensor)
        p = _lib.apply_ConvolutionLayer(self.pointer, input.pointer)
        return _Tensor(_c.c_void_p(p))

    def __del__(self):
        _lib.del_ConvolutionLayer(self.pointer)


class UpsamplingLayer:

    def __init__(self, data):
        if isinstance(data, _c.c_void_p):
            self.pointer = data
        elif isinstance(data, dict):
            settings = _json.dumps(data)
            self.pointer = _lib.new_UpsamplingLayer(settings.encode("utf-8"))
        else:
            raise TypeError(f"Unexpected type '{type(data)}'"
                            "in UpsamplingLayer's constructor")

    def __call__(self, input, concat):
        assert isinstance(input, _Tensor)
        p = _lib.apply_UpsamplingLayer(self.pointer, input.pointer,
                                       concat.pointer)
        return _Tensor(_c.c_void_p(p))

    def __del__(self):
        _lib.del_UpsamplingLayer(self.pointer)


class DenseLayer:

    def __init__(self, data):
        if isinstance(data, _c.c_void_p):
            self.pointer = data
        elif isinstance(data, dict):
            settings = _json.dumps(data)
            self.pointer = _lib.new_DenseLayer(settings.encode("utf-8"))
        else:
            raise TypeError(f"Unexpected type '{type(data)}'"
                            "in DenseLayer's constructor")

    def __call__(self, input):
        assert isinstance(input, _Tensor)
        p = _lib.apply_DenseLayer(self.pointer, input.pointer)
        return _Tensor(_c.c_void_p(p))

    def __del__(self):
        _lib.del_DenseLayer(self.pointer)



import torch
import torch.nn.functional as F
import numpy as np
from typing import Union, List, Tuple
from .layers import Linear, BatchNorm2d
from .layers import Encoder2d, OverallPool2d, FullyConnected
from .conversions import floats_to_tensor, tensor_to_index
from .conversions import categories_to_tensor, images_to_tensor
from .nn_decorators import nn_classifier


@nn_classifier
class ImageClassifier(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        pass

    def __init__(self, in_channels: int,
                 categories: List[str],
                 convolutions: Union[List[dict], List[List[dict]]],
                 pooling: List[Tuple[int, int]],
                 fully_connected: List[int] = [],
                 pooling_type: str = "Max",
                 padded: bool = True,
                 activation: str = "relu"):
        super().__init__()
        assert len(convolutions) == len(pooling)
        self.in_channels = in_channels
        self.categories = list(categories)
        self.input_norm = BatchNorm2d(in_channels)
        self.encoder = Encoder2d(in_channels, convolutions, pooling,
                                 pooling_type, padded=padded,
                                 activation=activation)
        self.overall_pool = OverallPool2d(pooling_type)
        self.fully_connected = FullyConnected(self.encoder.out_channels,
                                              fully_connected,
                                              activation=activation)
        self.output = Linear(self.fully_connected.out_features,
                             len(categories))

    def forward(self, X: torch.Tensor):
        X = self.input_norm(X)
        X = self.encoder(X)
        X = self.overall_pool(X)
        X = self.fully_connected(X)
        return self.output(X)

    def data_to_tensor(self, X: np.ndarray,
                       Y: Union[None, List[str]],
                       weights: Union[None, List[float]] = None
                       ) -> tuple:
        x = images_to_tensor(X, self.device)
        y = None if Y is None else categories_to_tensor(Y, self.categories,
                                                        self.device)
        w = None if weights is None else floats_to_tensor(weights)
        return x, y, w

    def tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        indexes = tensor_to_index(tensor)
        return [self.categories[i] for i in indexes]

    def loss(self, y_pred: torch.Tensor, y_target: torch.Tensor,
             weights: Union[torch.Tensor, None]):
        if weights is None:
            return F.cross_entropy(y_pred, y_target, weight=self.class_weights)
        else:
            return F.nll_loss(F.log_softmax(y_pred, dim=-1) * weights,
                              y_target, weight=self.class_weights)

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "shape in": self.shape_in,
                "categories": list(self.categories),
                "input norm": self.input_norm.dump,
                "convolutional": [c.dump for c in self.convolutional],
                "fully connected": [d.dump for d in self.fully_connected],
                "output": self.output.dump}



import torch
import torch.nn.functional as F


class Activation(torch.nn.Module):
    """A module applying an activation function to its input"""

    @classmethod
    def from_dump(cls, dump: dict) -> 'Activation':
        """returns an 'Activation' object from a dump"""
        assert dump["type"] == cls.__name__
        return cls(dump["function name"])

    def __init__(self, name: str = "relu"):
        """
        Parameters
        ----------
        name : str {"relu", "leaky_relu", "elu", "tanh", "sigmoid"}
            The name of the activation function to apply
        """
        super().__init__()
        self.function_name = name
        if name == "identity":
            self.function = lambda x: x
        elif hasattr(torch, name):
            self.function = getattr(torch, name)
        else:
            self.function = getattr(F, name)

    def forward(self, input: torch.Tensor):
        """Sequence of operations applied by the Module"""
        return self.function(input)

    @property
    def dump(self) -> dict:
        """returns a dump of the layer"""
        return {"type": type(self).__name__,
                "function name": self.function_name}
